# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/26_markdown.obsidian.personal.machine_learning.definition_identification.ipynb.

# %% auto 0
__all__ = ['definitions_in_text', 'definition_identification_data_from_note', 'gather_definition_identification_data']

# %% ../../../../../nbs/26_markdown.obsidian.personal.machine_learning.definition_identification.ipynb 2
import os 
from os import PathLike
from pathlib import Path
from typing import Union

import pandas as pd

from .....helper import current_time_formatted_to_minutes, definition_asterisk_indices
from ....markdown.file import MarkdownFile
from ..note_processing import process_standard_information_note
from ...vault import VaultNote

# %% ../../../../../nbs/26_markdown.obsidian.personal.machine_learning.definition_identification.ipynb 5
def definitions_in_text(
        text: str
        ) -> list[str]:
    """
    Return the list of str with the definitions in the text.
    """
    indices = definition_asterisk_indices(text)
    return [text[start+2:end-2] for start, end in indices]

# %% ../../../../../nbs/26_markdown.obsidian.personal.machine_learning.definition_identification.ipynb 7
# TODO: implement a measure to not get the definition identification data, e.g. by 
# detecting a `_auto/definition_identification` tag.
def definition_identification_data_from_note(
        note: VaultNote,
        vault: PathLike
        ) -> Union[dict[str, str], None]: # The keys to the dict are "Note name", "Raw text", "Definitions". However, `None` is returned if `note` does not exist.
    """Obtain definition identification data from the information note.

    """
    if not note.exists():
        return None
    mf = MarkdownFile.from_vault_note(note)
    mf = process_standard_information_note(
        mf, vault, remove_double_asterisks=False)
    mf_text = str(mf)
    definitions = definitions_in_text(mf_text)
    raw_text = mf_text.replace('**', '') 
    return {
        "Note name": note.name,
        "Raw text": raw_text,
        "Definitions": definitions}

# %% ../../../../../nbs/26_markdown.obsidian.personal.machine_learning.definition_identification.ipynb 9
def gather_definition_identification_data(
        vault: PathLike,
        notes: list[VaultNote]
        ) -> pd.DataFrame: # 
    """
    Return a `pandas.DataFrame` encapsulating the data of definition
    identifications.
    
    cf. `definition_identification_data_from_note`, which is the function
    with which the definition identification data is drawn.
    
    This function is mainly used in
    `append_to_definition_identification_database`.
    """
    definition_identification_data = [
        definition_identification_data_from_note(note, vault) for note in notes]
    definition_identification_data = [
        row for row in definition_identification_data
        if row is not None]
    current_time = current_time_formatted_to_minutes()
    for row in definition_identification_data:
        row['Time added'] = current_time
        row['Time modified'] = current_time
    return pd.DataFrame(definition_identification_data)

# %% ../../../../../nbs/26_markdown.obsidian.personal.machine_learning.definition_identification.ipynb 11
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb.

# %% auto 0
__all__ = ['index_note_for_reference', 'reference_directory', 'delete_reference_folder', 'copy_obsidian_vault_configs',
           'get_obsidian_vault_plugin_configs', 'modify_obsidian_vault_plugin_configs',
           'copy_obsidian_vault_configs_with_nice_modifications', 'setup_folder_for_new_reference',
           'get_index_notes_from_subdirectory', 'get_index_notes_from_index_note', 'reference_folders_in_vault',
           'files_in_reference_folder']

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 2
import glob
import json
import os
from os import PathLike
from pathlib import Path
import re
import shutil
from typing import Union, Optional
import warnings


from trouver.helper import (
    path_name_no_ext, alphabet_to_alphabet_group
)
from trouver.markdown.markdown.file import (
    MarkdownFile, MarkdownLineEnum
)
from trouver.markdown.markdown.heading import (
    heading_title
)
from trouver.markdown.obsidian.links import (
    ObsidianLink, LinkType, links_from_text
)
from .authors import find_author_file
from trouver.markdown.obsidian.personal.index_notes import ( 
    convert_title_to_folder_name
)
from trouver.markdown.obsidian.personal.notes import (
    notes_linked_in_note
)
from trouver.markdown.obsidian.personal.note_type import (
    type_of_note, PersonalNoteTypeEnum
)
from trouver.markdown.obsidian.vault import(
    VaultNote, all_note_paths_by_name, note_path_by_name,
    NoteDoesNotExistError, NoteNotUniqueError
)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 5
def index_note_for_reference(
        vault: PathLike, # The vault in which the reference folder resides.
        reference: Union[str, Path] # - The reference. Is either - a str, in which case the reference folder will be the folder containing the (unique) note of the name `_index_{reference}.md`, - or a `Path` object (not just a pathlike!) relative to `vault`, in which case the path will be the path to the reference folder. 
        ) -> VaultNote:
    """
    Returns the index note of the specified reference in the vault.
    
    Assumes that the reference folder has an index note named
    `_index_{reference_name}.md` and this note is the unique note in the vault
    with this filename.
    
    **Raises**

    - TypeError
        - If `reference` is not a str or PathLike.
    - NoteDoesNotExistError
        - If a note of the name `_index_{reference_name}.md` does not exist
        in the vault.
    """
    if (not isinstance(reference, str)
            and not isinstance(reference, PathLike)):
        raise TypeError(
            "Expected `reference` to be a str or a PathLike, but got"
            f" {type(reference)} instead.")
    if isinstance(reference, str):
        reference_name = reference
        index_note = VaultNote(vault, name=f'_index_{reference_name}')
    elif isinstance(reference, PathLike):
        reference_name = Path(reference).name
        index_note = VaultNote(
            vault, rel_path=Path(reference) / f'_index_{reference_name}.md')
    return index_note



# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 8
def reference_directory(
        vault: PathLike, # The vault in which the reference folder resides.
        reference: Union[str, Path] # - The reference. Is either - a str, in which case the reference folder will be the folder containing the (unique) note of the name `_index_{reference}.md`, - or a `Path` object (not just a pathlike!) relative to `vault`, in which case the path will be the path to the reference folder. 
        ) -> Path: # Relative to `vault`.
    """
    Returns the path to the reference directory in a vault.
    
    Assumes that the reference folder has an index note named
    `_index_{reference_name}.md`, this note is the unique note in the vault
    with this filename, and the cache in the `VaultNote` class for `vault` is
    updated.

    **Raises**

    - TypeError
        - If `reference` is not a str or PathLike.
    - 
    
    """
    index_note = index_note_for_reference(vault, reference)
    if index_note.exists(update_cache=False):
        return Path(index_note.path(relative=True)).parent
    else:
        raise NoteDoesNotExistError.from_note_name(index_note.name)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 14
# TODO: test delete reference and template files when reference folder does not exist.
# TODO: this method is long. make private helper methods
def delete_reference_folder(
        vault: PathLike, # The vault in which the reference folder resides.
        reference: Union[str, PathLike], # The reference to delete. Is either a str, in which case the folder to delete will be the folder containing the (unique) note of the name `_index_{reference}.md`, or a path relative to `vault`. 
        verbose: bool = True,
        confirm: bool = True # If `True`, prompts the user to confirm the deletion of the folder.
        ) -> None:
    """
    Deletes a reference folder along with the associated template note
    and the reference note, both of which are outside the reference folder.
    
    Assumes that
    - the reference folder, if it exists, has an index note named
    `_index_{reference_name}.md` and this note is the unique note in the vault
    with this filename.
    - the template note, if it exists, is named `_template_{reference_name}.md`
    and is the unique note in the vault with this filename.
    - the reference note, if it exists, is named `_reference_{reference_name}.md`
    and is the unique note in the vault with this filename.
    
    If the template/reference note for the reference is not unique, then the
    deletion does not proceed. On the other hand, even if a template/reference
    note does not exist, then the deletion proceeds.

    Note that links to notes in the reference folder are preserved.
    
    **Raises**

    - FileNotFoundError
        - If the specified reference folder does not exist.
    - NoteDoesNotExistError
        - If the index note for the reference folder does not exist in the
        vault.
    - NoteNotUniqueError
        - If the index note, template note, or reference note for the reference
        folder is not unique in the vault.
    """
    try: 
        reference_path = reference_directory(vault, reference)
        absolute_path = Path(vault) / reference_path
        if verbose:
            print(f"\nIdentified reference '{reference}' in the vault '{vault}' as"
                f" the folder '{absolute_path}'...")
        reference_name = reference_path.name
    except NoteDoesNotExistError:
        reference_path = None
        reference_name = path_name_no_ext(str(reference))
    
    try:
        template_note = VaultNote(vault, name=f'_template_{reference_name}')
    except NoteDoesNotExistError:
        template_note = None
    try:
        reference_note = VaultNote(vault, name=f'_reference_{reference_name}')
    except NoteDoesNotExistError:
        reference_note = None
        
    if confirm:
        input_msg = [f"Delete"]
        if reference_path:
            input_msg = [f"\n- all contents in the folder '{absolute_path}'"]
        if template_note:
            input_msg.append(f"\n- '{template_note.path()}'")
        if reference_note:
            input_msg.append(f"\n- '{reference_note.path()}'")
        input_msg.append(f"?\n[Y/(n)]")
        command = input(''.join(input_msg))
        delete = command == 'Y'
    else:
        delete = True

    if delete:
        if verbose:
            print("Deleting...")
        if reference_path:
            shutil.rmtree(absolute_path)
            # TODO: delete the reference folder in a way that updates the cache.
        if template_note:
            template_note.delete()
        if reference_note:
            reference_note.delete()
        if verbose:
            print(f"Deleted reference.\n")
    elif verbose:
        print(f"Aborting deleting reference.\n")



# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 18
def _make_reference_folder(
        vault: Path, location: PathLike, reference_name: str,
        reference_directory: PathLike, overwrite: Union[str, None], verbose: bool) -> None:
    """Makes a folder of a specified name in the specified directory.
    
    If `overwrite` is `'w'`, then the reference folder is assumed to not
    exist.

    This method is Intended for making a reference folder for an Obsidian
    vault.
    
    **Parameters**
    - vault - PathLike
        - The path to the Obsidian vault in which to make the reference folder.
    - location - PathLike
        - The directory of the parent of the new folder to be made, relative
        to `vault`. 
    - reference_name - str
        - The name of the reference to be created; the folder's name will be
        this string.
    - reference_directory - PathLike
        - Is `location / reference_name`; in particular, this is a path
        relative to `vault`.
    - overwrite - `'w'`, `'a'`, or `None`.
        - Specifies if and how to overwrite the reference folder if it already
        exists.
            - If `'w'`, then the reference folder is assumed to not exist, and
            the reference folder is created.
            - If `'a'`, then the reference folder may or may not exist, and the
            reference folder is created if it does not exist.
            - If `None`, then the reference folder may or may not exist. If the
            reference folder exists, then a `FileExistsError` is raised. If the
            reference folder does not exist, then it is created.
    - verbose - bool
        - If `True`, print messages.
        
    **Raises**
    - FileNotFoundError
        - If `location` does not exist as a path relative to `vault`.
    - FileExistsError
        - If the reference folder already exists and 1. overwrite is `'w'` or
        2. overwrite is `'None'`. The former case is expected to not happen.
    """
    if verbose:
        print(f"Attempting to create a the folder '{reference_directory}'"
              f" in the directory '{vault / location}.'")

    if not os.path.exists(vault / location):
        raise FileNotFoundError(
            f"Attempted to create a reference folder at {vault / reference_directory}"
            f", but the parent directory {vault / location} does not" 
            " exist.")
    if os.path.exists(vault / reference_directory):
        if overwrite in ['w', None]:
            raise FileExistsError(
                "Attempted to create a reference folder at"
                f" {reference_directory}, but this directory already exists.")
    else:
        os.mkdir(vault / reference_directory)


# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 20
def _chapter_titles(chapters: list[Union[str, list[str]]]) -> list[str]:
    """
    Return the list of chapters of a reference from a formatted list of chapters
    and sections.

    **Parameters**
    - chapters - list[Union[str, list[str]]]
        - A list whose items are str or list of str. If an item is a string, then the
        item is the title of a chapter of the reference. If an item is a list of string,
        then the item contains the title of the chapter of the reference as its index-0
        item, and the titles of the sections for the chapter.

    **Returns**
     - list[str]
    """
    return [chapter if isinstance(chapter, str) else chapter[0] for chapter in chapters]

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 22
def _make_index_file(
        vault: PathLike, reference_directory: PathLike, reference_name: str,
        chapters: list[Union[str, list[str]]]) -> None:
    """
    Create the index file for the reference folder.

    The newly created index file will have a bulleted list of Obsidian links to the
    index files of the chapters/sections of the reference.

    **Parameters**
    - vault - PathLike
    - reference_directory - PathLike
    - reference_name - str
    - chapters - list[Union[str, list[str]]]
        - A list whose items are str or list of str. If an item is a string, then the
        item is the title of a chapter of the reference. If an item is a list of string,
        then the item contains the title of the chapter of the reference as its index-0
        item, and the titles of the sections for the chapter.
    """
    index_note = VaultNote(vault, rel_path = reference_directory / f'_index_{reference_name}.md')
    index_note.create()
    chapter_titles = _chapter_titles(chapters)

    chapter_bullets = [f'- [[_index_{convert_title_to_folder_name(chapter_title)}]]'
                        for chapter_title in chapter_titles]
    mf = MarkdownFile.from_list(chapter_bullets)
    mf.write(index_note)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 24
def _make_chapter_folders_and_indices(
        chapters: list[Union[str, list[str]]],
        vault: PathLike,
        reference_directory: PathLike) -> None:
    chapter_titles = _chapter_titles(chapters)
    for chapter_title, chapter_sections in zip(chapter_titles, chapters):
        _make_single_chapter_folders(chapter_sections, chapter_title, vault, reference_directory)
        _make_single_chapter_index(chapter_sections, chapter_title, vault, reference_directory)

def _make_single_chapter_folders(
        chapter_sections: Union[str, list[str]],
        chapter_title: str,
        vault: Path,
        reference_directory: PathLike):
    chapter_title = convert_title_to_folder_name(chapter_title)
    os.mkdir(vault / reference_directory / chapter_title)

    if not isinstance(chapter_sections, list):  # i.e. chapter does not have sections
        return
    subchapter_titles = chapter_sections[1:]
    for subchapter_title in subchapter_titles:
        os.mkdir(vault / reference_directory / chapter_title /\
                    convert_title_to_folder_name(subchapter_title))


def _make_single_chapter_index(
        chapter_sections: Union[str, list[str]],
        chapter_title: str,
        vault: Path,
        reference_directory: PathLike):
    chapter_title = convert_title_to_folder_name(chapter_title)
    chapter_index_note = VaultNote(
        vault, rel_path=reference_directory / chapter_title\
        / f"_index_{chapter_title}.md")
    chapter_index_note.create()

    subchapter_titles = chapter_sections[1:]
    headings = [f'# {subchapter_title}\n\n\n'
                for subchapter_title in subchapter_titles]
    mf = MarkdownFile.from_list(headings)
    mf.write(chapter_index_note)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 26
def _make_reference_file(
        reference_name, references_folder,
        vault,
        authors, author_files) -> None:
    """
    The references folder has subfolders 'A-E', 'F-J', etc. each of which contains
    subfolders for each letter in the English alphabet, each of which in turn contains
    reference notes
    """
    if not os.path.exists(vault / references_folder):
        raise FileNotFoundError(
            f"References folder does not exist: {vault / references_folder}")

    if reference_name[0].isalpha():
        alphabet_group = alphabet_to_alphabet_group(reference_name[0])
        folder_to_make_reference_file = Path(references_folder)\
            / alphabet_group / reference_name[0].upper()
    else:
        folder_to_make_reference_file = Path(references_folder)
    reference_file_path = folder_to_make_reference_file\
        / f'_reference_{reference_name}.md'
    VaultNote(vault, rel_path=reference_file_path).create()
    # TODO author stuff

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 28
def _manifest_template_file(
        template_file_name: str,
        reference_name: str,
        vault: PathLike,
        authors: Union[str, list[str]])\
        -> MarkdownFile:
    """
    Return a `MarkdownFile` object based on a specified template file
    with some information about the reference added.

    **Parameters**
    - template_file_name: str
        - The template file's name. This is assumed to be unique in
        the vault.
    - reference_name: str
    - vault: PathLike
    - authors: Union[str, list[str]]

    **Returns**
    - MarkdownFile

    **Raises**
    - NoteNotUniqueError
        - If the template file's name is not unique in the vault.
    - NoteDoesNotExistError
        - If a note with the template file's name does not exist in the
        vault.
    """
    # TODO: sort authors by alphabetical order
    if isinstance(authors, str):
        authors = [authors]
    # TODO: get MarkdownFile by VaultNote
    # template_file_path = note_path_by_name(template_file_name, vault)
    # template_file = MarkdownFile.from_file(
    #     Path(vault) / template_file_path)
    template_file = MarkdownFile.from_vault_note(
        VaultNote(vault, name=template_file_name))
    embedding_link = ObsidianLink(is_embedded=True,
        file_name=f'_reference_{reference_name}',
        anchor=0, custom_text=0, link_type = LinkType.WIKILINK)
    template_file.add_line_in_section(
        title="References",
        line_dict={'type': MarkdownLineEnum.DEFAULT,
                   'line': f'{embedding_link.to_string()}\n'})

    last_line = template_file.pop_line()
    last_line['line'] = f'[^1]: {", ".join(authors)}, '
    template_file.add_line_to_end(last_line)

    # TODO: delete the below
    # ref_line_index = template_file.get_line_number_of_heading(
    #     title="References")
    # embedding_link = ObsidianLink(is_embedded=True,
    #     file_name=f'_reference_{reference_name}',
    #     anchor=0, custom_text=0, link_type = LinkType.WIKILINK)
    # template_file.insert_line(
    #     ref_line_index+1, {'type': MarkdownLineEnum.DEFAULT,
    #                        'line': f'{embedding_link.to_string()}\n'})

    # TODO: make sure the above line works and delete the below line
    # template_file.parts.insert(ref_line_index+1, {'type': MarkdownLineEnum.DEFAULT,
    #                        'line': f'{embedding_link.to_string()}\n'})
    template_file.add_tags([f'#_reference/{reference_name}'])
    return template_file

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 30
def _make_template_file(
        template_file_name,
        reference_name,
        vault,
        templates_folder,
        authors: Union[str, list[str]],
        make_second_template_file_in_reference_directory: bool,
        reference_directory: Path
        ) -> None:
    template_file = _manifest_template_file(
        template_file_name, reference_name, vault, authors)

    # TODO factor out; this repeats with code in _make_reference_file above.
    if not os.path.exists(vault / templates_folder):
        raise FileNotFoundError(
            f"References folder does not exist: {vault / templates_folder}")
    if reference_name[0].isalpha():
        alphabet_group = alphabet_to_alphabet_group(reference_name[0])
        folder_to_make_template_file = Path(templates_folder)\
            / alphabet_group / reference_name[0].upper()
    else:
        folder_to_make_template_file = Path(templates_folder)

    template_file_path = folder_to_make_template_file\
        / f'_template_{reference_name}.md'
    _create_template_note_at(vault, template_file_path, template_file)

    if make_second_template_file_in_reference_directory:
        second_template_file_path = reference_directory / f'_template_{reference_name}_2.md'
        _create_template_note_at(vault, second_template_file_path, template_file)


def _create_template_note_at(
        vault: Path,
        note_path: Path, # Relative to `vault`
        template_file: MarkdownFile):
    template_note = VaultNote(vault, rel_path=note_path)
    template_note.create()
    template_file.write(template_note)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 32
def _make_notation_index_file(
        reference_directory: PathLike,
        reference_name: str,
        vault: PathLike,
        notation_index_template_file_name: str,
        authors: list[str]) -> None:
    """Create the notation index file for the reference.

    The notation index file is named `_notation_{reference_name}.md`, and is
    located in the main directory of the new reference.

    **Parameters**
    - reference_directory - PathLike
        - The main directory of the reference. Relative to `vault`.
    - reference_name - str
    - vault - PathLike
    - notation_index_template_file_name - str
        - The template file whose contents will fill the newly created notation
        index file.
    
    **Raises**
    - NoteNotUniqueError
        - If the template file's name is not unique in the vault.
    - NoteDoesNotExistError
        - If a note with the template file's name does not exist in the
        vault.
    """
    template_file = _manifest_template_file(
        notation_index_template_file_name, reference_name, vault, authors)
    notation_index_file_path = Path(reference_directory)\
        / f'_notation_{reference_name}.md'
    notation_index_note = VaultNote(vault, rel_path=notation_index_file_path)
    notation_index_note.create()
    template_file.write(notation_index_note)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 34
# TODO: change "file" in these helper methods and their respective doccstrings to "note"
def _make_glossary_file(
        reference_directory: PathLike, reference_name: str, vault: PathLike,
        glossary_template_file_name: str) -> None:
    """Create the glossary file for the reference.

    The notation index file is named `_glossary_{reference_name}.md`, and is
    located in the main directory of the new reference.

    **Parameters**
    - reference_directory - PathLike
        - The main directory of the reference.
    - reference_name - str
    - vault - PathLike
    - glossary_template_file_name - str
        - The template file whose contents will fill the newly created
        glossary file.
    
    """
    template_note = VaultNote(vault, name=glossary_template_file_name)
    template_file = MarkdownFile.from_vault_note(template_note)
    glossary_file_path = reference_directory / f'_glossary_{reference_name}.md'
    glossary_note = VaultNote(vault, rel_path=glossary_file_path)
    glossary_note.create()
    template_file.write(glossary_note)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 36
def _make_temp_folder(
        reference_directory: PathLike, reference_name: str,
        vault: PathLike) -> None:
    """
    Make a folder for temporarily holding notes and a corresponding index file.

    **Parameters**
    - reference_directory - PathLike
    - reference_name - str
    - vault - PathLike

    """

    temp_directory = vault / reference_directory / '_temp'
    os.mkdir(temp_directory)
    # TODO: use vaultnote.create and markdownfile.write
    temp_file_path = reference_directory / '_temp' / f'_index_temp_{reference_name}.md'
    temp_note = VaultNote(vault, rel_path = temp_file_path)
    temp_note.create()

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 40
def copy_obsidian_vault_configs(
        vault: PathLike,
        reference_directory: PathLike, # The folder into which to copy the Obsidian configs. Relative to `vault`.
        configs_folder: PathLike, # The folder containing the Obsidian configs. This is either an absolute path or relative to the current working directory.
        dirs_exist_ok: bool = False, # If `dirs_exist_ok` is `False` and `reference_directory` already exists, then a FileExistsError is raised. If `dirs_exist_ok` is true, the copying operation will continue if it encounters existing directories, and files within the destination tree will be overwritten by corresponding files from the source tree. See also the [`shutil.copytree`](https://docs.python.org/3/library/shutil.html#shutil.copytree) function.
        ) -> None:
    """
    Copy the vault's Obsidian config files into the reference directory.

    **Raises**
    - `FileExistsError`
        - If `dirs_exist_ok` is `False` and `reference_directory` already exists.
    """
    shutil.copytree(
        configs_folder, vault / reference_directory / '.obsidian',
        dirs_exist_ok=dirs_exist_ok)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 44
def _obsidian_vault_plugin_configs_file(
        vault: PathLike,
        plugin_name: str, # The folder name of the Obsidian plugin. This can be found either in the `.obsidian` directory or in the `.obsidian/plugins` directory in the vault .
        plugin_is_core: bool #`True` if the plugin is a core Obsidian.md plugin. `False` if the plugin is a community plugin.
        ) -> Path:
    if plugin_is_core:
        return Path(vault) / '.obsidian' / f'{plugin_name}.json'
    else:
        return Path(vault) / '.obsidian' / 'plugins' / plugin_name / 'data.json'

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 46
def get_obsidian_vault_plugin_configs(
        vault: PathLike,
        plugin_name: str, # The folder name of the Obsidian plugin. This can be found in the `.obsidian` directory in the vault.
        plugin_is_core: bool #`True` if the plugin is a core Obsidian.md plugin. `False` if the plugin is a community plugin.
        ) -> dict[str, Union[str, int, float, bool, None, list, dict]]: # A json dict. 
    """Obtain the JSON object representing the `data.json` file of
    an Obsidian plugin.
    
    This function assumes that the the Obsidian plugin exists in `vault` in that the plugin
    has a `data.json` file.
    """
    with open(_obsidian_vault_plugin_configs_file(vault, plugin_name, plugin_is_core), 'r') as file:
        return json.load(file)
    

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 48
def modify_obsidian_vault_plugin_configs(
        vault: PathLike,
        plugin_name: str, # The folder name of the Obsidian plugin. This can be found in the `.obsidian` directory in the vault.
        plugin_is_core: bool, #`True` if the plugin is a core Obsidian.md plugin. `False` if the plugin is a community plugin.
        field: str, # The field to modify
        value: Union[str, int, float, bool, None, list, dict], # The JSON value to set the field as
        ) -> None:
    """Modify/set a top level field in an Obsidian vault plugin configs file.
    
    Assumes that the Obsidian vault plugins configs are indented by 2 spaces.

    Note that only top level values can be directly set by this function.

    """
    configs = get_obsidian_vault_plugin_configs(vault, plugin_name, plugin_is_core)
    configs[field] = value
    with open(_obsidian_vault_plugin_configs_file(vault, plugin_name, plugin_is_core), 'w') as file:
        json.dump(configs, file, indent=2)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 50
def _modify_fast_link_edit_and_templates(
        vault: PathLike,
        reference_name: str, # The value to change the `referenceName` field in the `fast-link-edit` plugin into.
        template_location: str = '/' # The value to change the `folder` field in the `templates` plugin into.
    ) -> None:
    """Modify the `fast-link-edit` and `templates` Obsidian plugins for a vault if
    each exists."""
    try:
        modify_obsidian_vault_plugin_configs(
            vault, 'fast-link-edit', False, 'referenceName', reference_name)
    except FileNotFoundError:
        warnings.warn(
            "Attempted to modify the the new reference folder's configuration file"
            " for the `fast-link-edit` plugin, but the file does not exist.",
            UserWarning)

    try:
        modify_obsidian_vault_plugin_configs(
            vault, 'templates', True, 'folder', template_location)
    except FileNotFoundError:
        warnings.warn(
            "Attempted to modify the the new reference folder's configuration file"
            " for the `templates` plugin, but the file does not exist.",
            UserWarning)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 52
def copy_obsidian_vault_configs_with_nice_modifications(
        vault: PathLike,
        reference_directory: PathLike, # The folder into which to copy the Obsidian configs. Relative to `vault`.
        configs_folder: PathLike, # The folder containing the Obsidian configs to copy. This is either an absolute path or relative to the current working directory.
        reference_name: str = None, # The name of the reference and the value to change the `referenceName` field in the `fast-link-edit` plugin into. If `None`, then the reference name should be obtained as the name of `reference_directory`
        template_location: str = '/', # The location of the template file(s) and the value to change the `folder` field in the `templates` plugin into.
        dirs_exist_ok: bool = False, # If `dirs_exist_ok` is `False` and `reference_directory` already exists, then a FileExistsError is raised. If `dirs_exist_ok` is true, the copying operation will continue if it encounters existing directories, and files within the destination tree will be overwritten by corresponding files from the source tree. See also the [`shutil.copytree`](https://docs.python.org/3/library/shutil.html#shutil.copytree) function.
        ) -> None:
    """
    Copy the vault's Obsidian config files into the reference directory and make some nice moodifications
    """
    copy_obsidian_vault_configs(
        vault, reference_directory, configs_folder, dirs_exist_ok=dirs_exist_ok)
    if reference_name is None:
        reference_name = path_name_no_ext(reference_directory)
    # Note that the config file that is being modified belongs to the "subvault",
    # i.e. the reference folder.
    _modify_fast_link_edit_and_templates(vault / reference_directory, reference_name, template_location)

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 56
# TODO: add reference link in index note of parent folder
# TODO: Make another template file in the reference folder
# TODO: When copying over configs, change some things.
# TODO: add verbose option. 
#     Done: Verbose in _make_reference_folder
# TODO: add type hinting in hidden methods and docstrings.
# TODO: make a method for deleting reference.
# TODO: enable overwrite existing 
# TODO: make a function to detect connections between references.
# TODO: Implement overwrite existing files/overwrite entire folder option.
# TODO: for some reason, glossary isn't being added.
# TODO: test add obsidian config folder
# TODO: make TODO file


def setup_folder_for_new_reference(
        reference_name: str, # The name of the reference to be created; the folder's name will be this string.
        location: PathLike, # The directory of the parent of the new folder to be made, relative to `vault`.
        authors: Union[str, list[str]], # Each str is the family name of each author.
        vault: PathLike, # The path to the Obsidian vault in which to make the reference folder.
        author_folder: PathLike = '_mathematicians', # The directory where the author files are stored in.  Relative to `vault`. 
        references_folder: PathLike = '_references', # The directory where the references files are stored in.  Relative to `vault`.
        templates_folder: PathLike = '_templates', # The directory where the template files are stored in.  Relative to `vault`.
        template_file_name: str = '_template_common', # The template file from which to base the template file of the new reference.
        notation_index_template_file_name: str = '_template_notation_index', # The template file from which to base the notation index file of the new reference.
        glossary_template_file_name: str = '_template_glossary', # The template file from which to base the glossary file of the new reference. Defaults to `'_template_glossary'`.
        chapters: Optional[list[Union[str, list[str]]]] = None, # A list where each item is either a str or a list of str.  If the item is a str, then it should be the title of a chapter named in the format `"1. {title}"`, `"2. {title}"`, or `"I. {title}"`, `"II. {title}"` etc. If the item is a list of str, then the 0th item should be the title of the chapter, formatted as in the previous sentence, and the other items should be titles of subchapters/subsections, also formatted in the same manner (e.g. if the subchapter is 7.2 of a book, then it should be `"2. {title}"`).  Defaults to `None`, in which case no chapters are specified and hence no chapter folders and indices are created.
        setup_temp_folder: bool = True, # If `True`, creates a `_temp` folder with an index file. This folder serves to house notes auto-created from LaTeX text files before moving them to their correct directories. Defaults to `True`.
        make_second_template_file_in_reference_directory: bool = True, # If `True`, creates a copy of the template note within the directory for the reference.
        copy_obsidian_configs: Optional[PathLike] = '.obsidian', # The folder relative to `vault` from which to copy obsidian configs.  If `None`, then no obsidian configs are copied to the reference folder. Defaults to `.obsidian`. 
        overwrite: Union[str, None] = None, # Specifies if and how to overwrite the reference folder if it already exists.  - If `'w'`, then deletes the contents of the existing reference folder, as well as the template and reference file before setting up the reference folder before creating the new reference folder.  - If `'a'`, then overwrites the contents of the reference folder, but does not remove existing files/folders.  - If `None`, then does not modify the existing reference folder and raises a `FileExistsError`.
        confirm_overwrite: bool = True, # Specifies whether or not to confirm the deletion of the reference folder if it already exists and if `overwrite` is `'w'`. Defaults to `True`.
        verbose: bool = False
        ) -> None:
    """Creates and sets up a new folder for a new reference.
    
    More specifically, the following are set up:
    
    1. The folder
    2. An index file
    3. Chapter folders and indices; the folder names are given by
      `convert_title_to_folder_name(<title_with_numbering>)` where
      `title_with_numbering` is the title of the chapter with its
      numbering if available, e.g. `1. Introduction`.
    4. Files in the `Mathematicians` folder, if applicable.
    5. A reference file
        - For embedding in the standard information notes for the reference
    6. A template file
    7. A glossary file
    8. A `#_meta/reference` tag for the new reference
    9. A notation index file
    10. Optionally, a `_temp` directory
    11. Optionally, the `./obsidian` config folder from the vault is
    copied.
    
    After they are setup, the user should
    1. Add content in the reference file.
    2. Modify the last line of the template file.
        
    **Raises**
    - FileExistsError
        - If `overwrite` is `None` and a folder of the name `reference_name`
        exists at `Path(vault) / location`.
    """
    if verbose:
        print("Setting up a new folder for a reference.")
        print(f"\tReference name: {reference_name}")
        print(f"\tVault: {vault}")
        print(f"\tLocation of new folder (relative to vault): {location}")
    vault = Path(vault)
    if not chapters:
        chapters = []
    reference_directory = Path(location) / reference_name

    if overwrite == 'w' and os.path.exists(vault / reference_directory):
        delete_reference_folder(
            vault, reference_directory, verbose=verbose,
            confirm=confirm_overwrite)

    _make_reference_folder(
        vault, location, reference_name, reference_directory,
        overwrite=overwrite, verbose=verbose)
    _make_index_file(vault, reference_directory, reference_name, chapters)
    _make_chapter_folders_and_indices(chapters, vault, reference_directory)

    # Look for/make mathematician files

    # TODO: `author_files` ultimately does not really do anything. Change this
    author_files = []
    # for author in authors:
    #     author_file = find_author_file(vault, author, author_folder)
    #     if author_file is None:
    #         print('')  # TODO
    #     author_files.append(author_file)

    _make_reference_file(
        reference_name, references_folder, vault,
        authors, author_files)
    _make_template_file(
        template_file_name, reference_name, vault,
        templates_folder, authors, make_second_template_file_in_reference_directory,
        reference_directory)
    _make_notation_index_file(
        reference_directory, reference_name, vault,
        notation_index_template_file_name, authors)
    _make_glossary_file(
        reference_directory, reference_name, vault,
        glossary_template_file_name)
    if setup_temp_folder:
        _make_temp_folder(reference_directory, reference_name, vault)
    if copy_obsidian_configs is not None:
        configs_folder = vault / copy_obsidian_configs
        # copy_obsidian_vault_configs(vault, reference_directory, configs_folder)
        copy_obsidian_vault_configs_with_nice_modifications(
            vault, reference_directory, configs_folder)
        # Modify 
    if verbose:
        print(f'Created a new reference folder at {location}.'\
            ' Make sure to update the reference file and'\
            ' The files for new mathematicians!')

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 62
def get_index_notes_from_subdirectory(
        vault: PathLike, # The path to the Obsidian vault directory.
        subdirectory: PathLike, # The path, relative to `vault` of the subdirectory of of the vault for the reference.
        main_index_note: bool = False, # If `True`, include the main index note for the reference. This index note should be in the directory specified by `subdirectory`.
        as_vault_notes: bool = True # If `True`, returns the ``VaultNote`` objects for the index notes.  Otherwise, returns the paths to the index notes as paths, represented by str.
        ) -> list[Union[str, VaultNote]]: # Either of the names of the index notes in the vault or of the index notes as VaultNote objects, depending on `as_vault_notes`.
    """Returns list of index note files for reference
    """
    vault = Path(vault)
    reference_directory = vault / subdirectory
    index_notes = glob.glob(f'{reference_directory}/**/_index_*.md', recursive=True)
    index_notes = [index_note for index_note in index_notes 
                   if Path(os.path.dirname(index_note)) != reference_directory / "_temp"]
    if not main_index_note:
        index_notes = [index_note for index_note in index_notes
                       if Path(os.path.dirname(index_note)) != reference_directory]
    index_notes = [os.path.relpath(index_note, vault) for index_note in index_notes]
    if as_vault_notes:
        index_notes = [VaultNote(vault, rel_path=index_note) for index_note in index_notes]
    return index_notes

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 64
# TODO: reformat
def get_index_notes_from_index_note(
        vault: PathLike, reference_name: str,
        as_vault_notes: bool = True) -> list[Union[str, VaultNote]]:
    """Returns the list of index notes for the reference in the order
    that they are listed in the reference's main index note.
    
    Does so by searching the "main" index note in the reference folder.
    Returns the index notes in order that they are listed in the main index
    note.
    
    Assumes that the reference folder is "formatted correctly". This includes
    the assumption that the main index note for the reference is the unique
    note named `'_index_{reference_name}'` in the vault.
    
    **Parameters**

    - `vault` - PathLike
        - The path to the Obsidian vault directory
    - `reference_name` - str
        - The name of the reference folder in the vault.
    - `as_vault_notes` - bool
        - If `True`, returns the ``VaultNote`` objects for the index notes.
        Otherwise, returns the paths to the index notes as paths, represented
        by str. Defaults to `True`.

    **Returns**
    - list[Union[str, VaultNote]]
        - Either of the names of the index notes in the vault or of the 
        index notes as VaultNote objects, depending on `as_vault_notes`.
    
    **See Also**
    - ``get_notes_from_index_note`` in
    ``markdown.obsidian.personal.index_notes``.
    """
    vault = Path(vault)
    main_index_note = VaultNote(vault, name=f'_index_{reference_name}')
    mf = MarkdownFile.from_vault_note(main_index_note)
    text = str(mf)
    links = links_from_text(text)
    index_notes = [link.file_name for link in links]
    if as_vault_notes:
        index_notes = [VaultNote(vault, name=index_note)
                       for index_note in index_notes]
    return index_notes

# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 67
def reference_folders_in_vault(vault: PathLike) -> dict[str, str]:
    """Returns a dict of reference folders in vault.
    
    **Parameters**
    - vault - PathLike
        - The path to the Obsidian vault directory.

    **Returns**
    - dict
        - The keys are str, each of which is the name of a reference.
        The values are path strings to the reference folder, relative to
        `vault`.
    """
    vault = Path(MATH_VAULT_LOCATION)
    reference_folders = {}
    index_files = vault.glob(f'**/_index*.md')
    for index_file in index_files:
        # print(index_file)
        directory = os.path.dirname(index_file)
        # print(directory)
        directory = os.path.relpath(directory, vault)
        # print(directory)
        # print(f'{directory}/_notation*.md')
        notation_notes = vault.glob(f'{directory}/_notation*.md')
        if list(notation_notes):
            reference_name = os.path.basename(directory)
            reference_folders[reference_name] = directory
            if not os.path.exists(vault / directory /\
                                  f'_index_{reference_name}.md'):
                warnings.warn(f'''The index file for this reference
                              seems misnamed: {reference_name}''')
            if not os.path.exists(vault / directory /\
                                 f'_notation_{reference_name}.md'):
                warnings.warn(f'''The notation file for this reference
                              seems misnamed: {reference_name}''')
    return reference_folders


# %% ../../../../nbs/10_markdown.obsidian.personal.reference.ipynb 70
# TODO: reformat
def files_in_reference_folder(
        vault: PathLike, reference: str, as_list: bool = False,
        search_index_notes: bool = False,
        exclude_notes_of_type: list[PersonalNoteTypeEnum] = None)\
        -> Union[list[str], dict[str, str]]:
    """Returns a dict or list of files in a reference folder of a vault.

    TODO: also implement a method to get the files as listed in index notes.

    **Parameters**
    - vault - PathLike
    - reference - str
        - Name of the reference folder in the vault. The folder must have
        the index note file `'_index_{reference}.md'`, and there must not
        be two index note files of the same name in the vault.
    - as_list - bool
        - If `True`, then returns a list. Returns a dict otherwise. Defaults
        to `False`.
    - search_index_notes - bool
        - If `True`, then gets the notes by looking at the reference's index
        notes and returns the notes as ordered by the index notes.
    - exclude_notes_of_type - list of 
            ``markdown.obsidian.personal.note_type.PersonalNoteTypeEnum``.
        - The notes of the types specified here will be excluded in the dict
        or list returned. Defaults to `None`, in which case
        
    **Returns**
    - list[str] or dict[str, str]
        - Either a list of path strings or a dict whose keys are str and values
        are path strings. The keys are the names of the files. The Paths are
        relative to `vault`.
    """
    if not exclude_notes_of_type:
        exclude_notes_of_type = []
    vault = Path(vault)
    if search_index_notes:
        return _files_in_reference_folder_via_index_notes(
            vault, reference, as_list, exclude_notes_of_type)
    else:
        return _files_in_reference_folder_via_glob(
            vault, reference, as_list, exclude_notes_of_type)
    
def _files_in_reference_folder_via_index_notes(
        vault: PathLike, reference: str, as_list: bool,
        exclude_notes_of_type: list[PersonalNoteTypeEnum])\
        -> Union[list[str], dict[str, str]]:
    # TODO make sure that notes_linked_in_note is ordered correctly.
    index_of_index = VaultNote(vault, name=f'_index_{reference}')
    index_notes = notes_linked_in_note(index_of_index, as_dict=False)
    file_list = []
    for index_note in index_notes:
        notes_in_index_note = notes_linked_in_note(index_note, as_dict=False)
        file_list.extend(notes_in_index_note)
    file_list = [note.rel_path for note in file_list
                 if type_of_note(note) not in exclude_notes_of_type]
    if as_list:
        return file_list
    else:
        return {os.path.basename(file): file for file in file_list}
    
def _files_in_reference_folder_via_glob(
        vault: PathLike, reference: str, as_list: bool,
        exclude_notes_of_type: list[PersonalNoteTypeEnum])\
        -> Union[list[str], dict[str, str]]:
    vn = VaultNote(vault, name=f'_index_{reference}')
    reference_folder = Path(os.path.dirname(vn.path()))
    files = reference_folder.glob('**/*.md')
    file_list = [os.path.relpath(Path(file), start=vault) for file in files] 
    vault_notes = [VaultNote(vault, rel_path=file) for file in file_list]
    vault_notes = [vn for vn in vault_notes
                 if type_of_note(vn) not in exclude_notes_of_type]
    file_list = [vn.rel_path for vn in vault_notes]
    if as_list:
        return file_list
    else:
        return {os.path.basename(file): file for file in file_list}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/13_markdown.obsidian.personal.notes.ipynb.

# %% auto 0
__all__ = ['notes_linked_in_note', 'notes_linked_in_notes_linked_in_note', 'note_has_tag']

# %% ../../../../nbs/13_markdown.obsidian.personal.notes.ipynb 2
from typing import Union

from trouver.markdown.markdown.file import (
    MarkdownFile
)
from trouver.markdown.obsidian.links import (
    links_from_text
)
from trouver.markdown.obsidian.vault import (
    VaultNote
)


# %% ../../../../nbs/13_markdown.obsidian.personal.notes.ipynb 5
def notes_linked_in_note(
        list_note: VaultNote, # The `VaultNote` in which to find the links.
        as_dict: bool = True # If `True`, returns a dict. Returns a list otherwise.
        ) -> Union[dict[str, VaultNote], list[VaultNote]]: # If dict, the keys are the names of the vault notes and the values are the `VaultNote` objects. If list, then the entries are the `VaultNote` objects.
    """Returns a list or dict of VaultNotes of notes linked by 
    a specified note.
    """
    text = list_note.text()
    links = links_from_text(text)
    if as_dict:
        return {link.file_name: VaultNote(list_note.vault, name=link.file_name) for link in links}
    else:
        return [VaultNote(list_note.vault, name=link.file_name) for link in links]


def notes_linked_in_notes_linked_in_note(
    list_note: VaultNote, # The `VaultNote` with links to notes containing the links to obtain.
    as_dict: bool = True # If `True`, returns a dict. Returns a list otherwise. 
    ) -> Union[dict[str, VaultNote], list[VaultNote]]: # If dict, the keys are the names of the vault notes and the values are the `VaultNote` objects. If list, then the entries are the `VaultNote` objects.
    """Returns a list or dict of VaultNotes of notes 
    linked by notes which are linked by a specified note.
    """
    linked_in_list_note = notes_linked_in_note(list_note, as_dict=False)
    notes_for_each_note = [notes_linked_in_note(note, as_dict=True) 
                           for note in linked_in_list_note]
    all_notes = {}
    for notes_for_note in notes_for_each_note:
        for name, note in notes_for_note.items():
            all_notes[name] = note
    if as_dict:
        return all_notes
    else:
        return [note for _, note in all_notes.items()]     

# %% ../../../../nbs/13_markdown.obsidian.personal.notes.ipynb 9
def note_has_tag(
        note: VaultNote,
        tag: str # Without the hashtag `'#'`.
        ) -> bool:
    """Returns `True` if the note has the specified tag in its
    yaml frontmatter meta.
    """
    mf = MarkdownFile.from_vault_note(note)
    return mf.has_tag(tag)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/24_markdown.obsidian.personal.machine_learning.notation_identification.ipynb.

# %% auto 0
__all__ = ['add_one_double_asts_to_line', 'notation_data_from_text', 'notation_data_from_note',
           'append_notation_data_to_database', 'automatically_mark_notations']

# %% ../../../../../nbs/24_markdown.obsidian.personal.machine_learning.notation_identification.ipynb 2
import os
from os import PathLike
from pathlib import Path
import pandas as pd
import re
from typing import Optional, Union
import warnings

from fastai.text.learner import TextLearner

from trouver.helper import (
    find_regex_in_text, latex_indices, notation_asterisk_indices,
    defs_and_notats_separations, replace_string_by_indices,
    current_time_formatted_to_minutes
)
from ....markdown.file import MarkdownFile, MarkdownLineEnum
from ..note_processing import process_standard_information_note
from .database_update import append_to_database
from trouver.markdown.obsidian.vault import(
    # all_note_paths_by_name, note_path_by_name,
    VaultNote, NoteDoesNotExistError
)


# %% ../../../../../nbs/24_markdown.obsidian.personal.machine_learning.notation_identification.ipynb 6
def add_one_double_asts_to_line(
        line: str, # The text to which to add the double asterisks `**`
        start: int, # The first double asterisks are added in between `line[start-1]` and `line[start]`.
        end: int # The second double asterisks are added in between `line[end-1]` and `line[end]`.
        ) -> str: # The str obtained from `line` by surrounding the substring `line[start:end]` with double asterisks.
    # TODO: rename to add_one_double_asts_to_line. Better yet, also
    # implement a function which adds multiple double asts.
    """
    Return `line` with only one double asterisks `**` surrounded text.
    
    Used in `_definition_data_from_line`
    """
    return f'{line[:start]}**{line[start:end]}**{line[end:]}'

# %% ../../../../../nbs/24_markdown.obsidian.personal.machine_learning.notation_identification.ipynb 8
def notation_data_from_text(
        with_double_asts: str # May or may not have double asterisks to signify definitions and notations
        ) -> tuple[str, list[tuple[int, int, bool]]]:
    """Extracts data on the locations of notations in a text with
    double asterisks.
    
    Used in `notation_data_from_note`

    **Returns**

    - tuple[str, list[tuple[int, int, bool]]]
        - The str is the str `no_double_asts`, which is the same as
        `with_double_asts`, except with the double asterisks removed.
        - Each list represents a data point for a LaTeX math-mode
          string in `no_double_asts`and consists of

            1. The indices `start, end` where the data point considers
               whether or not the LaTeX math-mode substring
               `line_no_double_asts[start:end]` is surrounded by
               double-asterisks (and hene is supposed to introduce a notation).

            2. A bool which is `True`, if the data-point represents a
               str with double-asterisks surrounding a notation and `False`
               otherwise.
    """
    defs_and_notats = defs_and_notats_separations(with_double_asts)
    only_indices = [(start, end) for start, end, _ in defs_and_notats]
    replace_with = [with_double_asts[start+2:end-2]
                    for start, end in only_indices]
    no_double_asts = replace_string_by_indices(
        with_double_asts, only_indices, replace_with)

    bold_indices_in_no_double_asts = [
        (start - 4*i, end - 4*i - 4, is_notat)
        for i, (start, end, is_notat) in enumerate(defs_and_notats)]

    notation_indices = [(start, end, True) for start, end, is_notat 
                        in bold_indices_in_no_double_asts if is_notat]

    notat_indices_in_no_double_asts = [
        (start, end) for start, end, _ in bold_indices_in_no_double_asts]
    all_latex_indices = latex_indices(no_double_asts)
    non_notat_indices = [tuppy for tuppy in all_latex_indices
                         if tuppy not in notat_indices_in_no_double_asts]
    non_notat_indices = [(start, end, False)
                         for start, end in non_notat_indices]
    
    return no_double_asts, notation_indices + non_notat_indices


# %% ../../../../../nbs/24_markdown.obsidian.personal.machine_learning.notation_identification.ipynb 10
def _notation_data_with_indices(
        note: VaultNote, vault: PathLike) -> tuple[
            MarkdownFile, list[tuple[int, int, bool]]]:
    r"""Obtain notation data from a note including the indices.
    
    Used in `notation_data_from_note`

    **Parameters**
    - note - VaultNote
    - vault - PathLike

    **Returns**
    - str, list[list[str, int, bool]]
        - The str is the str of the processed MarkdownFile except
        without double asterisks.
        - Each list consists of

            1. The indices `start, end` where the data point considers
            whether or not the substring `no_double_asts[start:end]`
            contains a notation.
            2. A bool that is `True` if the LaTeX text contains
            notation.
    """
    # TODO: test
    mf = process_standard_information_note(
        MarkdownFile.from_vault_note(note), vault,
        remove_double_asterisks = False)
    with_double_asts = str(mf)
    no_double_asts, data = notation_data_from_text(with_double_asts)
    return no_double_asts, data

# %% ../../../../../nbs/24_markdown.obsidian.personal.machine_learning.notation_identification.ipynb 11
def notation_data_from_note(
        note: VaultNote, vault: PathLike
        ) -> list[tuple[str, str, bool]]:
    # TODO: Implement the option to include multiple-lines in the data.
    """Obtain notation data from a note.

    Note that the lists of str might not be in any particular order.
    
    **Returns**

    - list[tuple[str, str, bool]]
        - Each list consists of 
            1. The name of `note`,
            2. The processed str of `note` with only a single double
            asterisk surrounded LaTeX text. Note that the processed str
            merges display math mode text into single lines, cf.
            `process_standard_information_note`.
            3. A bool that is `True` if the LaTeX text contains
            notation.
    """
    # TODO: treat '`$`` separately.
    no_double_asts, data = _notation_data_with_indices(note, vault)
    return [
        (note.name,
         add_one_double_asts_to_line(no_double_asts, start, end),
         is_notat) for start, end, is_notat in data]


# %% ../../../../../nbs/24_markdown.obsidian.personal.machine_learning.notation_identification.ipynb 17
def append_notation_data_to_database(
        vault: PathLike, # The vault from which the data is drawn
        file: PathLike,  # The path to a CSV file
        notes: list[VaultNote], # The notes to add to the database
        backup: bool = True # If `True`, makes a copy of `file` in the same directoy and with the same name, except with an added extension of `.bak`.
        ) -> None:
    to_turn_into_a_df = []
    current_time = current_time_formatted_to_minutes()
    for note in notes:
        notation_data_for_note = notation_data_from_note(note,  vault)
        for _, with_one_double_asts, is_notat in notation_data_for_note:
            row_dict = {
                'Time added': current_time,
                'Time modified': current_time,
                'Note name': note.name,
                'LaTeX in text': with_one_double_asts,
                'Is notation': is_notat
            }
            to_turn_into_a_df.append(row_dict)
    df = pd.DataFrame(to_turn_into_a_df)
    append_to_database(
        file, df,
        cols=['Time added', 'Time modified', 'Note name',
              'LaTeX in text', 'Is notation'],
        pivot_column='LaTeX in text',
        columns_to_update=['Time modified', 'Note name', 'Is notation'],
        backup=backup)
    
    

# %% ../../../../../nbs/24_markdown.obsidian.personal.machine_learning.notation_identification.ipynb 21
def automatically_mark_notations(
        vn: VaultNote, # The information note to which to mark notations.
        learn: TextLearner, # The ML model which predicts where notation notes should occur.  This is a classifier which takes as input a str with double asterisks surrounding LaTeX text. The model outputs whether or not the single double asterisk pair surrounds a LaTeX text with notation.
        create_notation_notes: bool = False, # If `True`, creates the notations notes for the predicted notations and links them to the 'See Also' sections of the information notes.
        reference_name: str = '' # The name of the reference that `vn` belongs to; this is only relevant when `create_notation_notes=True` so that the created notation notes have file names starting with the reference name.
        ) -> None:
    # TODO: before running this, make sure to warn or check that this
    # will change contents of files drasticall.
    # TODO: implement `overwrite` parameter
    """Predict and mark where notations occur in a note, and optionally
    create a notation note, and add the notation note to the `See Also`
    section of the note.

    Assumes that no double asterisks are already in the contents of `vn`.

    This function Removes links, headings, footnotes, etc.
    from the original note and merges multi-line display math mode LaTeX
    text into single lines. Use with caution.
    """
    no_double_asts, index_data = _notation_data_with_indices(vn, vn.vault)
    notations_to_add = _get_notation_indices_to_add(
        no_double_asts, index_data, learn)
    with_double_asts = no_double_asts
    for start, end in reversed(notations_to_add):
        with_double_asts = add_one_double_asts_to_line(
            with_double_asts, start, end)

    original_mf = MarkdownFile.from_vault_note(vn)
    _, end_metadata = original_mf.metadata_lines()
    see_also_line = original_mf.get_line_number_of_heading('See Also')
    original_mf.remove_lines(end_metadata + 1, see_also_line)
    original_mf.insert_line(end_metadata + 1, {
        'type': MarkdownLineEnum.HEADING, 'line': '# Topic[^1]'})
    original_mf.add_line_in_section('Topic[^1]', {
        'type': MarkdownLineEnum.DEFAULT, 'line': with_double_asts})
    original_mf.write(vn)

    # if create_notation_notes:
    #     make_notation_notes_from_double_asts(vn, vn.vault, reference_name)


def _get_notation_indices_to_add(
        no_double_asts: str, index_data: list[list[int, bool]],
        learn: TextLearner)\
            -> list[tuple[int]]:
    """Used in `automatically_add_notations`"""
    to_test = [add_one_double_asts_to_line(no_double_asts, start, end)
                       for start, end, is_notat in index_data if not is_notat]
    with learn.no_bar(), learn.no_logging():
        predictions = [learn.predict(one_double_ast)
                       for one_double_ast in to_test]
    notations_to_add = [
        (start, end) for (start, end, is_notat), prediction
        in zip(index_data, predictions) if is_notat or prediction[0] == 'True']
    notations_to_add.extend([
        (start, end) for (start, end, is_notat) in index_data if is_notat])
    notations_to_add.sort()
    return notations_to_add




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb.

# %% auto 0
__all__ = ['get_latex_in_original_from_parsed_notation_note_data', 'notation_summarization_data_from_note',
           'gather_notation_note_summaries', 'append_to_notation_note_summarization_database',
           'single_input_for_notation_summarization', 'append_column_for_single_text', 'fix_summary_formatting',
           'summarize_notation', 'append_summary_to_notation_note']

# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 3
import os
from os import PathLike
from pathlib import Path
import re
from typing import Optional, Union

import pandas as pd
from transformers import pipeline, pipelines

from .....helper import current_time_formatted_to_minutes
from ....markdown.file import MarkdownFile, MarkdownLineEnum
from .database_update import append_to_database
from ..note_processing import process_standard_information_note
from ..notation import parse_notation_note, main_of_notation
from ..note_type import note_is_of_type
from ...vault import VaultNote

# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 8
def get_latex_in_original_from_parsed_notation_note_data(
        metadata: dict[str],
        notation_str: str) -> str:
    """Return the string that should be considered the `latex_in_original`
    given part of the output of `parse_notation_note`

    `latex_in_original` is intended to either be 
    1. a substring of the main note in which the notation is introduced, 
       if available, or 
    2. the notation itself.

    If the notation note has a `latex_in_original` field in its YAML
    frontmatter metadata, then the (first) str there is the
    `latex_in_original`. 
    """
    if metadata is not None and 'latex_in_original' in metadata:
        latex_in_original = metadata['latex_in_original']
        if isinstance(latex_in_original, list):
            latex_in_original = latex_in_original[0]
    else:
        # latex_in_original = None 
        latex_in_original = notation_str
    return latex_in_original

# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 10
def _notation_note_has_auto_summary_tag(
        metadata: Union[dict[str], None]
        ) -> bool:
    """
    Return `True` if the metadata parsed from the notation note
    has an auto tag in its YAML frontmatter meta.

    This is a helper function of `notation_summarization_data_from_note`.
    """
    return metadata and 'tags' in metadata and '_auto/notation_summary' in metadata['tags']

# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 12
def notation_summarization_data_from_note(
        notation_note: VaultNote,
        vault: PathLike
        ) -> Union[dict[str, str], None]: # The keys to the dict are "Notation note name", "Notation", "Latex in oiriginal", "Summary", and "Main note name". However, `None` is returned instead of a `dict` if the notation note is determined to have not been summarized, if the main note of the notation note does not exist, or the notation note is marked with the `_auto/notation_summary` tag.
    """Obtain notation summzarization data from the notation note.

    `None` is returned instead of a `dict if

    1. the notation is determined to have not been (fully) summarized,
       - Currently, this is implemented as follows: a notation note
         is not fully summarized if it does not have any content
         beyond `<notation> denotes ` or if the notation note has
         `#_meta/TODO` marked in its content
         (except in the trailing bulleted list of notations used
          in the notation note)
    2. the main note of the notation not does not exist or
       is essentially empty
       - Essentially empty means that the
         `process_standard_information_note` function applied to the
         a `MarkdownFile` object of the main note yields a
         `MarkdownFile` object whose `__str__` method returns a
         string of only blank characters.
    3. the notation note is marked with an `_auto/notation_summary` tag
       in its YAML frontmatter meta.
       - If a notation note is marked with an `_auto/notation_summary`
         tag, then that means that its summary has been auto-generated
         by an ML model using the `append_summary_to_notation_note`
         function

    The notion of whether the notation "has been summarized" is not
    exactly implemented, but should be sufficient for gathering
    data.

    This function is mainly used in `gather_notation_note_summaries`.

    TODO: there are some notation notes drawing information from
    multiple notes; separate text in notation notes corresponding
    to each note.

    **Returns**

    - Union[dict[str, str], None]

        - If the notation note is determined to have been summarized
          (i.e. contains text beyond `<notation> denotes ` and does not
          have the `#_meta/TODO` tag) then the output is a `dict` whose
          key-value pairs are
            - `"Notation note name"` - The name of the notation note
            - `"Notation"` - The notation of the note
            - `"Latex in original"' - The entry of the `latex_in_original`
              field of the note if available, cf. `make_a_notation_note`
            - `"Summary"` - The summary of the notation.
            - `"Main note name"` - The name of the main note of the
              notation note
            - `"Processed main note contents"` - The processed contents of the
              main note
        - Otherwise, the output is `None.
    """
    metadata, notation_str, main_of_notation, main_mf, _ = parse_notation_note(notation_note, vault)
    if not _notation_has_been_summarized(main_mf):
        return None
    if _notation_note_has_auto_summary_tag(metadata):
        return None

    latex_in_original = get_latex_in_original_from_parsed_notation_note_data(
        metadata, notation_str)

    process_standard_information_note(main_mf, vault)
    processed_summary = str(main_mf)
    if len(processed_summary.strip()) == 0:
        return None

    if main_of_notation is None:
        return None
    main_note = VaultNote(vault, name=main_of_notation)
    if not main_note.exists():
        return None
    main_note_mf = MarkdownFile.from_vault_note(main_note)
    process_standard_information_note(main_note_mf, vault)

    return {"Notation note name": notation_note.name,
            "Notation": notation_str,
            "Latex in original": latex_in_original,
            "Summary": processed_summary,
            "Main note name": main_of_notation,
            "Processed main note contents": str(main_note_mf)}


def _notation_has_been_summarized(
        main_mf: VaultNote
        ) -> bool:
    text = str(main_mf).strip()
    return len(text) > 0 and '#_meta/TODO' not in text



# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 22
def gather_notation_note_summaries(
        vault: PathLike,
        notes: list[VaultNote]
        ) -> pd.DataFrame: # Has columns `Time added`, `Time modified`, `Notation note name`, `Notation`, `Latex in original`, 'Summary', 'Main note name', and 'Processed main note contents'. 
    """
    Return a `pandas.DataFrame` encapsulating the data of notation note
    summaries.

    cf. `notation_summarization_data_from_note`, which is the function
    from which the notation note summaries are drawn.

    This function is mainly used in
    `append_to_notation_note_summarization_database`.
    """
    summary_data = [
        notation_summarization_data_from_note(note, vault) for note in notes]
    summary_data = [row for row in summary_data if row is not None]
    current_time = current_time_formatted_to_minutes()
    for row in summary_data:
        row['Time added'] = current_time
        row['Time modified'] = current_time
    return pd.DataFrame(summary_data)
    

# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 25
def append_to_notation_note_summarization_database(
        vault: PathLike, # The vault freom which the data is drawn
        file: PathLike, # The path to a CSV file
        notes: list[VaultNote], # the notation notes to consider adding to the database. 
        backup: bool = True # If `True`, makes a copy of `file` in the same directory and with the same name, except with an added extension of `.bak`.
        ) -> None:
    """
    Either create a `csv` file containing data for notation note
    summarization or append to an existing `csv` file.

    The columns of the database file are as follows:

    - `Time added` - The time when the row was added.
    - `Time modified` - The time when the labels of the row 
    - `Notation note name` - The name of the note from which the data for the row
      was derived.
    - 'Notation' - The notation which is being summarized
    - 'Latex in original' - The entry of the `latex_in_original` field of the
      note if available, cf. `make_a_notation_note`
    - `"Summary"` - The summary of the notation.
    - `"Main note name"` - The name of the main note of the
      notation note
    - `"Processed main note contents"` - The processed contents of the
      main note

    All timestamps are in UTC time and specify time to minutes
    (i.e. no seconds/microseconds).
    
    TODO: implement updating rows and rewrite the next paragraph to
    accurately reflect the implementation. I would like the 'Notation', 'Latex in original',
    'Summary', 'processed main note contents' to be the "pivot_cols"

    If a "new" note has the same processed content as a pre-existing
    note and anything is different about the "new" note, then update
    the row of the existing note. In particular, the following are updated:
    - Time modified (set to current time)
    - Notation (overwritten)
    - Latex in original (overwritten)
    - Summary (overwritten)
    - Main note name (overwritten)
    - Processed main note contents (overwritten)
    
    This method assumes that all the processed content in the
    CSV file are all distinct if the CSV file exists.
    """
    if not notes:
        return
    file = Path(file)
    df = pd.read_csv(file) if os.path.exists(file) else None
    # start_ID_from = max_ID(df) + 1 if not df is None else 1
    new_df = gather_notation_note_summaries(vault, notes)
    if new_df.empty:
        return
    cols = [
        'Time added', 'Time modified', 'Notation note name',
        'Notation', 'Latex in original', 'Summary', 'Main note name',
        'Processed main note contents']
    cols_to_update = [
      'Time modified', 'Processed main note contents', 'Notation', 'Latex in original', 'Summary', 'Main note name']
    append_to_database(
        file, new_df, cols, 'Notation note name', cols_to_update, backup)
# TODO: think about whether the 'Notation note name' column would make for an
# appropriate pivot.

# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 34
def single_input_for_notation_summarization(
        main_note_content: str, # The mathematical text that introduces the notation and from which to summarize a notation.
        latex_in_original: str, # A substring in main_note_content which is a latex string in which the notation is introduced.
        ) -> str:
    """
    Format an input for a
    `transformers.pipelines.text2text_generation.SummarizationPipeline`
    object to summarize a notation introduced in a mathematical text.

    Note that this function is used to format data used to train/validate
    the summarization model within the `SummarizationPipeline`.
    """
    return f"{main_note_content}\n\nlatex_in_original: {latex_in_original}"

# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 38
# TODO: I wonder if I should also keep text that doesn't take 
# Latex in original but rather the notation itself.
def append_column_for_single_text(
        df: pd.DataFrame # Assumed to be structured just as a dataframe of a CSV file created/modified by append_to_notation_note_summarization_database``
        ) -> None:
    """Append a column `"Single text"` to the notation note summarization
    DataFrame to represent the input into an ML model as a single text
    """
    single_text_column = df.apply(
        lambda row: single_input_for_notation_summarization(
            row["Processed main note contents"], row["Latex in original"]),
        axis=1)
    df["Single text"] = single_text_column

# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 44
def fix_summary_formatting(
        summary: str
        ) -> str:
    """Fix some latex formatting issues in a summarized text
    """
    summary = summary.replace(r'\ ', '\\')
    summary = summary.replace(r'{ ', r'{')
    summary = summary.replace(r' }', r'}')
    summary, _ = re.subn(r'\$\s*([^\$]+?)\s*\$', r'$\1$', summary)
    # TODO: do $ <latex_string> $ into $<latex_stinrg>$
    # TODO: if the replacement of r'\ ' by '\\' happesn to
    # make `\` stick to the previous chunk of things
    # (e.g. r'd\in\mathbb{Z}_{\geq 0}`, then give it some
    # space, e.g. r'd \in \mathbb{Z}_{\geq 0}'.
    return summary



# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 48
def summarize_notation(
        main_content: str,
        latex_in_original: str,
        summarizer: pipelines.text2text_generation.SummarizationPipeline,
        fix_formatting: bool = True, # If `True`, run `fix_summary_formatting` on `summarizer`'s summary before retuning it.
        ) -> str:
    """Summarize a notation introduced in a mathematical text using
    a huggingface pipeline.

    Assumes that `main_content` is a mathematical text introducing
    a notation and that `latex_in_original` is a substring of
    `main_content` in which a notation is introduced.
    """
    summarizer_output = summarizer(
        single_input_for_notation_summarization(main_content, latex_in_original))
    summary = summarizer_output[0]['summary_text']
    if fix_formatting:
        summary = fix_summary_formatting(summary)
    return summary

# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 50
def _escape_latex_in_original_in_metadata(notation_mf: MarkdownFile):
    """Escape the `latex_in_original` field in the metadata.
    
    Assumes that `notation_mf` is a MarkdownFile object derived
    from a notation note.
    """
    metadata = notation_mf.metadata()
    if not 'latex_in_original' in metadata:
        return
    latex_in_original = metadata['latex_in_original']
    latex_in_original = [
        _enquote(raw_notation.replace('\\', '\\\\'))
        for raw_notation in latex_in_original]
    metadata['latex_in_original'] = latex_in_original
    notation_mf.replace_metadata(metadata)



# %% ../../../../../nbs/25_markdown.obsidian.personal.machine_learning.notation_summarization.ipynb 52
def append_summary_to_notation_note(
        notation_note: VaultNote,
        vault: PathLike,
        summarizer: pipelines.text2text_generation.SummarizationPipeline, # Contains an ML model which summarizes the notation, see `summarize_notation` function.
        main_note: Optional[VaultNote] = None, # The main note from which the notation comes from. If this is `None`, then the `main_note` is obtained via the `main_of_notation` function.
        overwrite_previous_autogenerated_summary: bool = False # If `True`, overwrite previously autogenerated summaries
    ) -> None:
    """Summarize a notation introduced in a mathematical text
    using a huggingface pipeline and append said summarization to
    `notation_note`.

    If `notation_note` does not have a YAML frontmatter meta or
    does not have a `latex_in_original` field in its YAML frontmatter
    meta, then the actual notation is used as the `latex_in_original`.

    The following describes the circumstances under which the
    summarization does not happen:

    - If `main_note` is `None` and no main note of `notation_note` can
    be determined via the `main_of_notation` function.
    - If `overwrite_previous_autogenerated_summary` is `False` and
    `notation_note` has the `_auto/notation_summary` tag in its
    YAML frontmatter meta (if available).
    - If `notation_note` does not have a YAML frontmatter meta or 
    the `_auto/notation_summary` tag is not present in the
    YAML frontmatter meta and `notation_note` has nontrivial
    content (i.e. its content is not merely of the form
    `$<notation>$ [[<link_to_main_note|denotes]]`).

    If an auto-generated summary is appended, then this function
    adds an `_auto/notation_summary` tag to the notation note's
    YAML frontmatter, if not already present.
    This tag is intended to be used as follows:

    - The presence of this tag tells the reader that the summary
      has been autogenerated by this function.
    - The `notation_summarization_data_from_note` function (and
      by extension, the `gather_notation_note_summaries` and
      `append_to_notation_note_summarization_database` functions)
      avoids gathering notation summarization data from notation
      notes marked with this tag
    - The owner of the `Obsidian.md` vault can manually make
      modifications to the notation note if necessary and remove this
      tag to indicate that the summary is appropriate to be added
      to the note summarization database.
    """
    metadata, notation_str, main_note_name,\
        notation_note_content_mf, _\
        = parse_notation_note(notation_note, vault)
    if main_note is None and main_note_name is not None:
        main_note = VaultNote(vault, name=main_note_name)
        # main_note = main_of_notation(notation_note, as_note=True)

    if main_note is None:
        return
    main_mf = MarkdownFile.from_vault_note(main_note)
    process_standard_information_note(main_mf, vault)
    if (not overwrite_previous_autogenerated_summary 
            and _metadata_dict_has_auto_notation_summary_tag(metadata)):
        return
    notation_note_content = str(notation_note_content_mf)
    if (not _metadata_dict_has_auto_notation_summary_tag(metadata)
            and len(notation_note_content.strip()) != 0):
        # TODO: warn that the notation note already had
        # contents, so no new ones were added.
        return
    
    summary = _get_summary(metadata, notation_str, main_mf, summarizer)
    _write_summary_to_notation_note(notation_note, summary)


def _get_summary(
        metadata, notation_str, main_mf, summarizer) -> str:
    """
    This is a helper function of `append_summary_to_notation_note`.
    """
    latex_in_original = get_latex_in_original_from_parsed_notation_note_data(
        metadata, notation_str)
    summary = summarize_notation(main_mf, latex_in_original, summarizer)
    return summary


def _write_summary_to_notation_note(
        notation_note: VaultNote, summary: str, ) -> None:
    """
    This is a helper function of `append_summary_to_notation_note`.

    
    """
    notation_note_mf = MarkdownFile.from_vault_note(notation_note)
    notation_note_mf.parts[-1]['line'] += summary
    notation_note_mf.add_tags(['_auto/notation_summary'])

    # _escape_latex_in_original_in_metadata(notation_note_mf)
    notation_note_mf.write(notation_note)


def _metadata_dict_has_auto_notation_summary_tag(
        metadata: Union[dict, None]):
    """
    This is a helper function of `append_summary_to_notation_note`.
    """
    return (metadata is not None
            and 'tags' in metadata
            and '_auto/notation_summary' in metadata['tags'])
    


    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/06_markdown.obsidian.links.ipynb.

# %% ../../../nbs/06_markdown.obsidian.links.ipynb 3
from __future__ import annotations
from deprecated import deprecated
from enum import Enum
from pathlib import Path
import re
from trouver.helper import (
    find_regex_in_text, text_from_file
)
from trouver.markdown.obsidian.vault import (
    all_paths_to_notes_in_vault, VaultNote, NoteDoesNotExistError
)
from typing import Union


# %% auto 0
__all__ = ['WIKILINK_PATTERN', 'EMBEDDED_WIKILINK_PATTERN', 'WIKILINK_CAPTURE_PATTERN', 'MARKDOWNLINK_PATTERN',
           'EMBEDDED_MARKDOWNLINK_PATERN', 'MARKDOWNLINK_CAPTURE_PATTERN', 'EMBEDDED_PATTERN',
           'find_links_in_markdown_text', 'LinkFormatError', 'LinkType', 'ObsidianLink', 'links_from_text',
           'remove_links_from_text', 'all_custom_text_for_links_in_vault', 'all_links_in_vault']

# %% ../../../nbs/06_markdown.obsidian.links.ipynb 4
# TODO Make it so that these patterns don't capture latex code
# WIKILINK_PATTERN = r'!?\[\[.*?\]\]' 
WIKILINK_PATTERN = r'!?\[\[[^\]]+\]\]'
EMBEDDED_WIKILINK_PATTERN = r'!\[\[[^\]]+\]\]'
WIKILINK_CAPTURE_PATTERN = r'!?\[\[([^#\|]*?)(#(.*?))?(\|(.*?))?\]\]'

# Note that MARKDOWNLINK_PATTERN captures whitespace characters in its link, even though Obsidian
# does not. This is implmeneted to find if any misformats in the Obsidian Markdown files.
MARKDOWNLINK_PATTERN = r'!?\[[^\]]+\]\([^)]+\)'  
EMBEDDED_MARKDOWNLINK_PATERN = r'!\[[^\]]+\]\([^)]+\)'
MARKDOWNLINK_CAPTURE_PATTERN = r'!?\[([^\]]*)\]\(([^)#]+)(#([^)]+))?\)'


EMBEDDED_PATTERN = f'{EMBEDDED_WIKILINK_PATTERN}|{EMBEDDED_MARKDOWNLINK_PATERN}'
# MARKDOWNLINK_CAPTURE = r'!?\[([^\]]+)\]\(([^)#])+(#[^)]+)?\)'

# %% ../../../nbs/06_markdown.obsidian.links.ipynb 7
def find_links_in_markdown_text(
        text: str
        ) -> list[tuple]: # Each tuple is of the form `(a,b)` where `text[a:b]` is an obsidian internal link.
    # TODO: rename this function, say to link_ranges_in_text, 
    # because it is confusing when there is a links_from_text function below.
    """Return ranges in the markdown text string
    where internal links occur.

    **See Also**

    - `links_from_text`
    """
    regex = f'{WIKILINK_PATTERN}|{MARKDOWNLINK_PATTERN}'
    return find_regex_in_text(text, pattern=regex)


# %% ../../../nbs/06_markdown.obsidian.links.ipynb 12
class LinkFormatError(Exception):
    """Error that is raised when a string cannot be parsed as an
    `ObsidianLink` object.
    
    **Attribute**

    - `text` - `str`
    """
    def __init__(self, text):
        self.text = text
        super().__init__(f'Obsidian Markdown link is not formatted properly: {text}')

# %% ../../../nbs/06_markdown.obsidian.links.ipynb 13
class LinkType(Enum):
    """An Enumeration indicating whether an `ObsidianLink` object is a
    Wikilink or a Markdown-style link.

    Enumerates `LinkType.WIKILINK` and `LinkType.MARKDOWN`.
    """
    # See https://www.markdownguide.org/basic-syntax/
    WIKILINK = 0
    MARKDOWN = 1  
    # For Markdown links, use %20 to encode spaces in the link, e.g.
    # [asdf](localization_of_a_module#Localization%20of%20a%20module%201)
    # Links to the header `"Localization of a module 1"` in the file
    # localization_of_a_module



# %% ../../../nbs/06_markdown.obsidian.links.ipynb 14
# TODO: implment equality, copy
class ObsidianLink:
    """Object representing an obsidian link
    
    **Attributes**

    - `is_embedded` - `bool`
        - Whether or not the link is embedded.
    - `file_name` - `str`, or `-1`
        - The destination of the link. It is either 
        
          1. The Obsidian-vault-recognized name of the file that the link
          points to. It can be a path relative to the Obsidian vault path 
          without the file extension (.md), 
          2. an external link, such as a URL, or
          3. -1, in which case the object represents a generic link pointing
          to any file (this is for generating regex).
          
          Note that if `file_name` is the empty string, then the link is a
          link to the same file

    - `anchor` - `str`, `0`, or `-1`
        - The title of the header of the anchor in the destination that the
        link points to or the ID to the markdown block link (preceded by a
        carat `^`). If 0, then the `ObsidianLink` object represents a link
        without an anchor. If -1, then the object represents a generic link
        with or without an anchor (this is for generating regex).
    - `custom_text` - `str`, `0`, or `-1`
        - The custom text of the link. Is `None` if no such text is specified.
        If 0, then the `ObsidianLink` object represents an internal link
        without custom text. If -1, then the object represents a generic
        internal link of any custom text (this is for generating regex).
    - `link_type` - `LinkType`
        - If `LinkType.WIKILINK`, then the str should be of the format
        `'[[<Obsidian-vault-recognized-name>(#anchor)?(|custom_text)]]'` 
        (The question marks here indicate optional components). Otherwise,
        the str should be a more standard Markdown link. Defaults to
        `LinkType.WIKILINK`.
    
    **Parameters**

    - is_embedded - bool
    - file_name - str or `None`
        - If `None`, set `self.file_name` to `-1`.
    - anchor - str or `None`
    - custom_text - str or `None`
    - link_type - `LinkType`
    """
    
    def __init__(
            self, is_embedded: bool, file_name: Union[str, int],
            anchor: Union[str, int], custom_text: Union[str, int],
            link_type: LinkType = LinkType.WIKILINK):
        self.is_embedded = is_embedded
        self.file_name = file_name
        self.anchor = anchor
        self.custom_text = custom_text
        self.link_type = link_type


    @staticmethod
    def from_text(text: str) -> ObsidianLink:
        """Return an ObsidianLink object from text.
                
        **Raises**

        - InteralLinkFormatError
            - If `text` is not properly formatted as an Obsidian internal link.
        """
        is_embedded = text.startswith("!")
        regex_object = re.compile(WIKILINK_CAPTURE_PATTERN)
        matches = regex_object.match(text)
        if matches:
            file_name = matches.group(1)
            anchor = matches.group(3)
            custom_text = matches.group(5)
            link_type = LinkType.WIKILINK
        else:
            regex_object = re.compile(MARKDOWNLINK_CAPTURE_PATTERN)
            matches = regex_object.match(text)
            if not matches:
                raise LinkFormatError(text)
            file_name = matches.group(2).replace('%20', ' ')
            anchor = matches.group(4)
            if anchor:
                anchor = anchor.replace('%20', ' ')
            custom_text = matches.group(1)
            link_type = LinkType.MARKDOWN
        if anchor is None:
            anchor = 0
        if custom_text is None:
            custom_text = 0
        return ObsidianLink(is_embedded, file_name, anchor, custom_text, link_type)

    def _parse_text_as_wikilink(text: str):
        """
        Return details about the link `text` if `text` is a Wikilink or `None` if
        `text is not a Wikilink.

        This is a helper method for `from_text`.
        """
        # TODO
        return

    def to_regex(self
            )-> str: # Represents a regex.
        """Return the regex for that this `ObsidianLink` object represents.

        Assumes that `self.file_name`, `self.anchor`, and `self.custom_text` are
        regex-formatted strings, e.g. if `self.custom_text` is `denotes?`, then the
        outputted regex-pattern matches links whose custom text is either `denote`
        or `denotes`.

        If neither `self.file_name`, `self.anchor` nor `self.custom_text` is `-1`,
        then the regex will in fact be a concrete string.
        """
        embedding = '!' if self.is_embedded else ''

        if type(self.file_name) == str:
            filing = self.file_name
        else:  # self.file_name == -1
            filing = r'([^#\|]*)?'
        
        if type(self.anchor) == str:
            anchoring = f'#{self.anchor}'
        elif self.anchor == 0:
            anchoring = ''
        else:  # self.anchor == -1
            anchoring = '(#(.*?))?'
          
        if type(self.custom_text) == str and self.link_type == LinkType.WIKILINK:
            customing = fr'\|{self.custom_text}'
        elif type(self.custom_text) == str and self.link_type == LinkType.MARKDOWN:
            customing = self.custom_text
        elif self.custom_text == 0:
            customing = ''
        else:  # self.custom == -1
            if self.link_type == LinkType.MARKDOWN:
                customing = fr'(.*?)?'
            else:
                customing = fr'(\|(.*?))?'

        if self.link_type == LinkType.WIKILINK:
            return fr'{embedding}\[\[{filing}{anchoring}{customing}\]\]'
        else:
            # Markdown links format whitespace with '%20'
            filing = filing.replace(' ' , '%20')  
            anchoring = anchoring.replace(' ', '%20')
            return fr'{embedding}\[{customing}\]\({filing}{anchoring}\)'
    
    def __str__(self) -> str:
        # TODO: Choose what to do about | vs. \|.
        return self.to_string()

    def to_string(self
            ) -> str: # The string for the link
        """
        Return the string for the link if it is concrete.
 
        **Raises**

        - ValueError
            - If `self.file_name`, `self.anchor` or `self.custom_text`
            is -1, i.e. ambiguously represents an anchor or custom text.
        """
        if self.is_abstract():
            raise ValueError(
                f'The ObsidianLink object is abstract.'
            )
        assert (self.anchor != -1 and self.custom_text != -1
                and self.file_name != -1)
        embedding = '!' if self.is_embedded else ''

        if type(self.anchor) == str:
            anchoring = f'#{self.anchor}'
        else:  # self.anchor == 0
            anchoring = ''
          
        if type(self.custom_text) == str:
            if self.link_type == LinkType.WIKILINK:
                customing = fr'|{self.custom_text}'
            else:
                customing = self.custom_text
        else:  # self.custom_text == 0:
            customing = ''
        
        if self.link_type == LinkType.WIKILINK:
            return f'{embedding}[[{self.file_name}{anchoring}{customing}]]'
        else:
            # Markdown links format whitespace with '%20'
            file_name = self.file_name.replace(' ' , '%20')  
            anchoring = anchoring.replace(' ', '%20')
            return fr'{embedding}[{customing}]({file_name}{anchoring})'
    
    def convert_link_type(
            self,
            link_type: LinkType
            ) -> ObsidianLink:
        """
        Return an equivalent Link object which has the specified
        `LinkType`.
        """
        # TODO
        return
    
    def displayed_text(self
            ) -> str: # The displayed text
        # TODO: implement error if any of the attributes is -1
        """Returns the displayed text of this link.
        
        `self.file_name`, `self.custom_text` and `self.anchor` are
        assumed to be not `-1`.
        """
        if self.custom_text:
            return self.custom_text
        else:
            if not self.anchor:
                return self.file_name
            else:
                return f'{self.file_name} > {self.anchor}'

    def is_abstract(self) -> bool:
        """
        Return `True` if self is abstract, i.e. file_name, anchor,
        or custom_text is `-1`.
        """
        return self.anchor == -1 or self.file_name == -1 or self.anchor == -1
    

# %% ../../../nbs/06_markdown.obsidian.links.ipynb 51
def links_from_text(
        text: str
        ) -> list[ObsidianLink]: # The `ObsidianLink` objects are ordered by appearance.
    """
    Return a list of `ObsidianLink` objects corresponding to links
    found in the text.
    """
    ranges = find_links_in_markdown_text(text)
    link_strs = [text[start:end] for start, end in ranges]
    return [ObsidianLink.from_text(link_str) for link_str in link_strs]

# %% ../../../nbs/06_markdown.obsidian.links.ipynb 55
def remove_links_from_text(
        text: str,
        exclude: list[ObsidianLink] = None, # A list of `ObsidianLink` objects of links to not be removed.
        remove_embedded_note_links: bool = False # If `True`, remove links to embedded notes as well. Note that embedded links are replaced by their "display" text in the same manner as non-embedded links and are not replaced the content of the embedding. If `False`, does not modify embedded note links.
        ) -> str:
    """
    Return a text with all Obsidian links removed and replaced with
    the display texts of the links.
    """
    if not exclude:
        exclude = []
    exclude_patterns = [re.compile(exclude_link.to_regex())
                        for exclude_link in exclude]
    link_indices = find_links_in_markdown_text(text)
    new_text = text
    for start, end in reversed(link_indices):
        if _do_not_remove_link(text[start:end], exclude_patterns):
            continue
        link_object = ObsidianLink.from_text(text[start:end])
        if link_object.is_embedded and not remove_embedded_note_links:
            continue
        replace_with = link_object.displayed_text()
        # link_object.custom_text if link_object.custom_text else link_object.file_name
        new_text = new_text[0:start] + replace_with + new_text[end:]
    return new_text


def _do_not_remove_link(text: str, exclude_patterns: list[re.Pattern]) -> bool:
    """
    Return `True` if text fully matches any of `re.Pattern`'s in `exclude_patterns`.

    This is a helper method for `remove_links_from_text`. 
    """
    for exclude_pattern in exclude_patterns:
        if exclude_pattern.fullmatch(text):
            return True
    return False

# %% ../../../nbs/06_markdown.obsidian.links.ipynb 64
def all_custom_text_for_links_in_vault(
        note: VaultNote, # The note to find the custom text for.
        vault: PathLike, # The path to the Obsidian vault directory
        anchor: Union[str, int] = -1 # The anchor in the note to find the custom text for. If 0, then returns only the custom texts for internal links without anchors. If -1, then returns the custom texts for all of the internal links of the specified note.
        ) -> dict[Union[str, int], str]: # Each key is the custom text used and each value is a set of paths relative to `vault` to the notes where the custom text is used.  The key `0` means that no custom text is used.
    """
    Return all custom text used in the Obsidian vault for the specified
    note and anchor.
    """
    vault = Path(vault)
    name = note.name
    wiki_gen = ObsidianLink(
        is_embedded=False, file_name=name, anchor=anchor, custom_text=-1,
        link_type=LinkType.WIKILINK)
    wiki_regex = wiki_gen.to_regex()
    mark_gen = ObsidianLink(
        is_embedded=False, file_name=name, anchor=anchor, custom_text=-1,
        link_type=LinkType.MARKDOWN)
    mark_regex = mark_gen.to_regex()
    regex = f'{wiki_regex}|{mark_regex}'
    all_note_paths = all_paths_to_notes_in_vault(vault)
    # I am creating a generator as opposed to a list so that the program
    # does not store the contents of all the text files.
    texts = ((path, text_from_file(vault / path, encoding='utf8'))
             for path in all_note_paths)
    regex_object = re.compile(regex)
    custom_text_usage = {}
    for path, text in texts:
        custom_texts = _custom_text_for_links_in_text(text, regex_object)
        for custom_text in custom_texts:
            if custom_text not in custom_text_usage:
                custom_text_usage[custom_text] = set()
            custom_text_usage[custom_text].add(path)
    return custom_text_usage


def _custom_text_for_links_in_text(
        text: str, regex_object: re.Pattern) -> list[str]:
    """Helper function for finding custom text for links in a single file.
    
    **Parameters**
    - text - str
    - regex_object - re.Pattern
        - An object returned with the `re.compile` function. Matches
        wikilinks and markdown links.
        
    **Returns**
    - list[str]
    """
    link_matches = regex_object.finditer(text)
    match_ranges = [match.span() for match in link_matches]
    links = [ObsidianLink.from_text(text[start:end])
             for start, end in match_ranges]
    return [link.custom_text for link in links]

# %% ../../../nbs/06_markdown.obsidian.links.ipynb 66
def all_links_in_vault(
        vault: PathLike, backlinks: bool = False, 
        multiplicities: bool = False) -> dict[str, list[str]]:
    """Returns a dict keeping track of which notes contain 
    links to which notes.
    
    TODO: currently, the regex capture regex code. Change them
    so that this does not happen.
    
    **Parameters**
    - vault - PathLike
    - backlinks - bool
        - If `True`, then keeps track of the links in each note.
        If `False`, then keeps track of the backlinks in each note,
        i.e. which other notes link to each note. Defaults to `False`.
    - multiplicties - bool
        - If `True`, then keeps track of multiple links to the same note
        for each note. Otherwise, only keeps track of whether or not
        a note links to a(nother) note. Defaults to `False`.
  
    **Returns**
    - dict[str, list[str]]
        - Each key is a str, referring to the name of a note. 
        Each corresponding value is a list. The list contains 
        the names of all the notes which the key note references to if
        `backlinks` is `True`, and contains the names of all the notes which
        reference the key note if `backlinks` is `False`. The list can contain
        multiple occurrences of the same note if `multiplicities` is `True`.
    """
    vault = Path(vault)
    all_note_paths = all_paths_to_notes_in_vault(vault, as_dict=True)
    links = {}
    for name, path in all_note_paths.items():
        text = text_from_file(vault / path, encoding='utf8')
        try:
            links_in_note = links_from_text(text)
        except LinkFormatError:
            print(name)
        if not multiplicities:
            links_in_note = set(links_in_note)
        if backlinks:
            if name not in links:
                links[name] = []
            for link in links_in_note:
                if link.file_name not in links:
                    links[link.file_name] = []
                links[link.file_name].append(name)
        else:
            if name not in links:
                links[name] = []
            links[name] += [link.file_name for link in links_in_note]
    return links



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb.

# %% auto 0
__all__ = ['subsections_listed_in_index_note', 'subsection_folders', 'get_alphanumeric', 'correspond_headings_with_folder',
           'information_notes_linked_in_index_note', 'move_information_notes_to_correct_folder',
           'move_information_notes_to_correct_folder_for_all_indices', 'convert_title_to_folder_name',
           'convert_heading_to_folder_name', 'make_folders_from_index_note_headers', 'get_notes_from_index_note',
           'add_link_in_index_note_after_note_link']

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 3
import glob
import os
from os import PathLike
from pathlib import Path
import re
from typing import Union

from natsort import natsorted
from pathvalidate import sanitize_filename

from trouver.markdown.markdown.file import (
    MarkdownFile, MarkdownLineEnum
)
from ...markdown.heading import heading_title
from trouver.markdown.obsidian.links import (
    find_links_in_markdown_text, ObsidianLink, links_from_text
)
from trouver.markdown.obsidian.vault import (
    VaultNote, note_name_unique, note_path_by_name
)

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 6
def subsections_listed_in_index_note(
        index_note: Union[VaultNote, str], # The index note
        vault: PathLike
        ) -> dict[Union[int, str], [dict, str]]: # The keys are 1. line numbers and 2. `'title'`. The values are dict and str (the blank str if root node), respectively.
    """
    Return subsections/subchapters as listed in the index note

    **See Also**
    
    - The `get_headings_tree` function of the `MarkdownFile` class.
    """
    vault = Path(vault)
    if isinstance(index_note, str):
        index_note = VaultNote(vault, name=index_note)
    mf_file = MarkdownFile.from_vault_note(index_note)
    return mf_file.get_headings_tree()

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 8
def subsection_folders(
        index_note: Union[VaultNote, str], # The index note
        vault: PathLike,
        output_type: str, # `'absolute_path'`, `'relative_path'`, or `'name'`
        ) -> list[str]: # List of immediate subdirectories in the directory containing the index note.
    """
    Return subdirectories corresponding to subsections/subchapters, i.e.
    the folders in the same directory as the index note.

    The folders are arranged in the order specified by `natsorted`.
    """
    vault = Path(vault)
    if isinstance(index_note, str):
        index_note = VaultNote(vault, name=index_note)
    parent_directory = (vault / index_note.rel_path).parent
    # print(str(parent_directory))
    glob_result = natsorted(glob.glob(str(parent_directory) + '/**/'))
    if output_type == 'absolute_path':
        return glob_result
    elif output_type == 'relative_path':
        return [str(Path(dir).relative_to(vault)) for dir in glob_result]
    elif output_type == 'name':
        return [Path(dir).name for dir in glob_result]

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 12
def get_alphanumeric(
        title: str, # The title of either a folder or a heading. Must start with an alphanumeric.
        title_type: str # Either `folder` or `heading`.
        ) -> str: # An alphabet or a numeric (arabic or roman)
    """
    Get the alphanumeric of a title of either a folder or a heading
    in an index noteh.

    Assumes that each folder is titled
    `'{alphanumeric}_{folder_title}'` and each heading is titled
    `'{alphanumeric}. {heading_title}'`
    """
    assert title_type in ['folder', 'heading']
    if title_type == 'folder':
        return re.sub(r'(.*?)_.*' , r'\1', title)
    else:
        return re.sub(r'(.*?)\. .*', r'\1', title)
    


# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 14
def correspond_headings_with_folder(
        index_note: VaultNote,
        vault: PathLike,
        include_non_heading: bool = True # If `True`, and if there is text before any heading, then treat such text as being under a "blank" heading.
        ) -> dict[str, str]:
    """
    Return tuples of corresponding headings in an index note
    with folder names.
    
    Assumes that each folder is titled
    `'{alphanumeric}_{folder_title}'` and each heading is titled
    `'{alphanumeric}. {heading_title}'`
    
    **Returns**
    - dict[str, str]
        - Each key is a str indexing the headings and folders. The keys
        are usually alphanumerics (arabic or roman), depending on the
        numbering system of chapters/sections of the reference/text.
        The values are tuples `(folder_title, heading_title)` without 
        the alphanumeric. For the blank heading, the key/index, the folder title,
        and the heading title are all the empty str.
    """
    index = MarkdownFile.from_vault_note(index_note)
    headings = index.get_headings(levels=1)
    headings = [heading_title(heading) for heading in headings]
    folders = subsection_folders(index_note, vault, output_type='name')
    correspond_dict = {get_alphanumeric(heading, 'heading'): (heading, folder)
                       for heading, folder in zip(headings, folders)}
    # TODO do a better job at the conditional below; 
    # for example, consider the start of the text blank if it's just empty lines with spaces.
    if (include_non_heading and index.parts
            and index.parts[0]['type'] != MarkdownLineEnum.HEADING):
        correspond_dict[''] = ('', '')
    return correspond_dict
    

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 19
def information_notes_linked_in_index_note(
        index_note: VaultNote, # The note indexing the information notes.
        vault: PathLike,
        hints: list[PathLike] = None # Hints on where the information notes are likely to be found at.  Each path is relative to `vault` and points to a directory. Defaults to `None`.
        ) -> dict[str, list[VaultNote]]: # Each key is the index for the heading (usually either an alphanumerical or a roman numerical). Each value is a list of the information notes linked in the index note.
    """Find information notes to be moved to the correct folder.
    
    Current implementation just looks at level 1 headings.
    This function is used in `move_information_notes_to_correct_folder`.
    Assumes that all notes in the vault have unique names.
    """
    parent_folder = os.path.dirname(index_note.rel_path)
    headings_folders = correspond_headings_with_folder(index_note, vault)
    mf = MarkdownFile.from_vault_note(index_note)
    headings_text = mf.get_headings_and_text(levels=1, include_start=True)
    headings_text = {heading_title(heading): text for heading, text
                     in headings_text.items()}
    text_under_headings = {heading_index: headings_text[heading] 
                            for heading_index, (heading, _) in headings_folders.items()}
    links_by_headings = {heading_index:links_from_text(text) for
                         heading_index, text in text_under_headings.items()}
    note_names_by_headings = {heading_index:[il.file_name for il in links] 
                              for heading_index, links in links_by_headings.items()}
    # Find notes by headings, but also pass the folder corresponding to the heading
    # as a hint of where to find the note for speedup, in case the note is 
    # already at the right place.
    folders_by_index = {heading_index: Path(vault) / parent_folder / heading_folder 
                        for heading_index, (_, heading_folder) in headings_folders.items()}
    if not hints:
        hints = []
    notes_by_headings = {heading_index: [VaultNote(vault, name=nn, hints=hints+[folders_by_index[heading_index]]) 
                                         for nn in note_names]
                         for heading_index, note_names in note_names_by_headings.items()}
    return notes_by_headings
    

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 21
def move_information_notes_to_correct_folder(
        index_note: VaultNote,
        vault: PathLike,
        hints: list[PathLike] = None # Hints on where the information notes are likely to be found at.  Each path is relative to `vault` and points to a directory. Defaults to `None`.
        ) -> None:
    """Moves the information notes indexed by `index_note` to the correct folder.

    The "correct folder" is a folder in the same directory as `index_note`
    corresponding to the heading under which the information note is indexed.
    The current implementation just looks at level 1 headings.
    """
    parent_folder = os.path.dirname(index_note.path(relative=True))
    linked_notes = information_notes_linked_in_index_note(index_note, vault, hints)
    headings_folders = correspond_headings_with_folder(index_note, vault)
    for heading_index, notes in linked_notes.items():
        _move_notes_under_heading(heading_index, notes,
                                  parent_folder, headings_folders)


def _move_notes_under_heading(
        heading_index, notes: list[VaultNote], parent_folder, headings_folders):
    destination_folder = headings_folders[heading_index][1]
    for note in notes:
        note_folder = os.path.dirname(note.rel_path)
        if destination_folder == note_folder:
            continue
        note.move_to_folder(Path(parent_folder) / destination_folder)


# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 23
def move_information_notes_to_correct_folder_for_all_indices(
        index_of_index_notes: VaultNote, # The index note indexing other index notes; `index_of_index_notes` is intended to be an index note for an entire reference whereas the index notes are intended to correspond to chapters/sections in the reference.
        vault: PathLike,
        hints: list[PathLike] = [] # Hints on where the information notes are likely to be found at.  Each path is relative to `vault` and points to a directory.
        ) -> None:
    """
    Moves the information notes for all index notes belonging to the reference as
    specified by `index_of_index_notes`.
    """
    index_of_index_file = MarkdownFile.from_vault_note(index_of_index_notes)
    text = str(index_of_index_file)
    index_files = find_links_in_markdown_text(text)
    index_files = [ObsidianLink.from_text(text[start:end])
                   for start, end in index_files]
    for link in index_files:
        index_note_name = link.file_name
        index_note = VaultNote(vault, name=index_note_name)
        move_information_notes_to_correct_folder(
            index_note, vault, hints=hints+[index_note.rel_path])

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 33
def convert_title_to_folder_name(title: str) -> str:
    # TODO: remove left/right
    """
    Returns a folder name for the given string, e.g. replaces spaces
    with underscore.
    
    **Parameters**
    - `title` - str
    
    **Returns**
    - str
    """
    characters_to_remove = [
        '.', '\'', '$', '(', ')', '{', '}', ':', '?', '!', ',', '#', '%', '&',
        '\\', '<', '>', '*', '?', '/', '"', '@', '+', '`', '|', '=', '[', ']',
        'mathscr', 'mathbf', 'mathrm', ]
    characters_to_turn_to_underscore = [' ', '-', '^']
    title = title.strip()
    title = title.lower()
    for character in characters_to_remove:
        title = title.replace(character, '')
    for character in characters_to_turn_to_underscore:
        title = title.replace(character, '_')
    title = sanitize_filename(title)
    return title

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 35
def convert_heading_to_folder_name(
        heading: str # Matches regex `\# (\w+?)\. (.*?)`
        ) -> str:
    """Converts a heading to a valid name for a folder.
    
    TODO Might not work correctly.

    **Parameters**
    - heading: str
        
    """
    regex_match = re.match('\# (\w+?)\. (.*)', heading)
    try:
        alphanumeric = regex_match.group(1)
        title = regex_match.group(2)
    except AttributeError:
        raise ValueError(
            f"`convert_heading_to_folder_name` unsuccessfully attempted"
            f" to match the following: {heading}")
        #print(heading)
    return f'{alphanumeric}_{convert_title_to_folder_name(title)}'    

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 37
def make_folders_from_index_note_headers(
        index_note: VaultNote
        ) -> None:
    """
    Make folders in the same directory as index note whose names
    are the titles of the headers of the index note.

    The headers of the index note must match the regex pattern `\# (\w+?)\. (.*?)`.
    """
    mfile = MarkdownFile.from_vault_note(index_note)
    headings = mfile.get_headings_by_line_number(levels=1)
    pattern = re.compile('\# (\w+?)\. (.*?)')
    folder_names = [convert_heading_to_folder_name(heading)
                    for _, heading in headings.items() if heading]
    directory = Path(os.path.dirname(index_note.path()))
    for folder_name in folder_names:
        try:
            os.mkdir(directory / folder_name)
        except OSError as error:
            pass

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 40
# TODO: do an example of the `include_embedded_notes` paramtere.
def get_notes_from_index_note(
        vault: PathLike, # The path to the Obsidian vault directory
        index_note: VaultNote, # The VaultNote object for the index note.
        as_vault_notes: bool = True, # If `True`, returns the ``VaultNote`` objects for the index notes. Otherwise, returns the names of these notes 
        include_embedded_notes: bool = False # If `True`, include in the list the embedded notes. Defaults to `False`.
        ) -> list[Union[str, VaultNote]]: # Either of the names of the index notes in the vault or of the index notes as VaultNote objects, depending on `as_vault_notes`.
    """Returns the list of notes listed in the index note in the order that
    they are listed in.
    
    Asssumes that the index note is "formatted correctly".
    
    **See Also**
    - ``get_index_notes_from_index_note`` in 
    ``markdown.obsidian.personal.reference``.
    """
    vault = Path(vault)
    text = index_note.text()
    links = links_from_text(text)
    if not include_embedded_notes:
        links = [link for link in links if not link.is_embedded]
    index_notes = [link.file_name for link in links]
    if as_vault_notes:
        index_notes = [VaultNote(vault, name=index_note)
                       for index_note in index_notes]
    return index_notes

# %% ../../../../nbs/12_markdown.obsidian.personal.index_notes.ipynb 46
def add_link_in_index_note_after_note_link(
        index_note: VaultNote,
        note_to_add_link_after: VaultNote,
        link_to_add: ObsidianLink) -> None:
    """
    Adds a link in the index note.

    The link is added after the first link to `note_to_add_link_after`. 
    If no link to `note_to_add_link_after` is found, then a link is added
    at the end.
    """
    link_to_find = ObsidianLink(False, note_to_add_link_after.name, -1, -1)
    pattern = link_to_find.to_regex()
    mf = MarkdownFile.from_vault_note(index_note)
    for i, part in enumerate(mf.parts):
        if re.search(pattern, part['line']):
            break
    mf.insert_line(i + 1, {'type': MarkdownLineEnum.UNORDERED_LIST,
                           'line': f'- {link_to_add.to_string()}'})
    mf.write(index_note)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb.

# %% auto 0
__all__ = ['MAX_NOTE_NAME_LENGTH', 'SPECIAL_CHARACTERS', 'replaceable_groups', 'REPLACEABLES', 'parse_notation_note',
           'notation_in_note', 'main_of_notation', 'notation_str_in_a_standard_information_note',
           'notation_notes_linked_in_see_also_section', 'notations_and_main_notes',
           'notation_note_is_linked_in_see_also_section', 'add_notation_note_to_see_also',
           'add_missing_notation_links_to_information_notes', 'notations_to_add_in_index',
           'index_notation_note_formatted_entry', 'make_a_notation_note', 'make_notation_notes_from_double_asts',
           'regex_from_latex', 'regex_from_notation_note']

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 2
from os import PathLike
from pathlib import Path
import re
from typing import Optional, Union
import warnings

from multiset import Multiset
from pylatexenc.latexwalker import LatexNode, LatexMacroNode, LatexWalker, LatexGroupNode, LatexCharsNode

from ....helper import notation_asterisk_indices, latex_to_path_accepted_string
from ...markdown.file import MarkdownFile, MarkdownLineEnum
from trouver.markdown.obsidian.links import (
    find_links_in_markdown_text, LinkType, ObsidianLink, MARKDOWNLINK_PATTERN, WIKILINK_PATTERN
)
from .information_notes import bulleted_links_of_type_in_section
from trouver.markdown.obsidian.personal.note_type import (
    PersonalNoteTypeEnum, assert_note_is_of_type, note_is_of_type
)
from ..vault import VaultNote, NotePathIsNotIdentifiedError, NoteDoesNotExistError

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 6
def _main_of_notation_from_text(
        file_text: str # Text of notation note
        ) -> Union[str, VaultNote, None]: # The name main information note that `notation_note` comes from. Returns `None` if `notation_note` does not come from such a note.
    """Return the name of the note from which the notation comes from.

    Helper function for `parse_notation_string`.
    """
    if '%%' in file_text and 'main: ' in file_text:
        return None

    # First, get rid of the notation, i.e. the first latex math mode string
    match = re.search(fr'\$.+?\$\s+', file_text) 
    if match is None:
        raise ValueError(
            'There seems to be a formatting error in a notation note'
            ' and the notation has not been identified. The following is the'
            f' content of the notation note:\n{file_text}')
    start, end = match.span()
    file_text = file_text[end:]
    
    link_locations = find_links_in_markdown_text(file_text)
    if not link_locations:
        return None
    start, end = link_locations[0]
    link_str = file_text[start:end]
    link_parse = ObsidianLink.from_text(link_str)
    main_note_name = link_parse.file_name
    return main_note_name

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 11
def _divide_bulleted_list_mf_at_end(
        mf: MarkdownFile
        ) -> tuple[MarkdownFile, Union[MarkdownFile, None]]: # The first MarkdownFile contains the main content. The second MarkdonwFile contains the bulleted list at the end; if no such bulleted list exists, then this is None.
    """Divide a `MarkdownFile` for a notation note into two MarkdownFiles, one
    of the main content and the other for the trailing bulleted list of links
    for notations used in the notation note.

    Assumes that the bulleted list is formatted correctly
    (i.e. each line is of the form `- [<notation>](<link>)`)

    Helper function for `parse_notation_note`.
    """
    main_parts = mf.parts.copy()
    trailing_parts = []
    for part in reversed(mf.parts):
        if part['type'] == MarkdownLineEnum.BLANK_LINE:
            main_parts.pop()
            continue
        if not _part_is_unordered_list_and_is_of_markdownstyle_link(part):
            break
        last_part = main_parts.pop() # Should be the same as `part`
        trailing_parts.insert(0, last_part)
    
    if trailing_parts:
        bulleted_list_mf = MarkdownFile(trailing_parts)
    else:
        bulleted_list_mf = None
    return MarkdownFile(main_parts), bulleted_list_mf

def _part_is_unordered_list_and_is_of_markdownstyle_link(
        part: dict[str, Union[str, MarkdownLineEnum]]
        ) -> bool:
    """
    
    Helper function for `_divide_bulleted_list_mf_at_end`
    """
    if part['type'] != MarkdownLineEnum.UNORDERED_LIST:
        return False
    if not part['line'].startswith('- '):
        return False
    if not re.match(MARKDOWNLINK_PATTERN, part['line'][2:]):
        return False
    return True
    

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 13
def parse_notation_note(
        notation_note: Union[str, VaultNote],
        vault: Optional[PathLike] = None # The vault If `None`, then uses th
        ) -> tuple[Union[dict, None], Union[str, None], str, MarkdownFile,
                   Union[MarkdownFile, None]]:
    """Parse information from the notation note.

    **Returns**

    - tuple[Union[dict, None], str, ObsidianLink, MarkdownFile, MarkdownFile]
        - The first entry is the YAML frontmatter meta, if available.
        - The second entry is the notation string
        - The third entry is the name of the "main note" of the notation note. This is usual
          the linked note in the link `[[<linked_note>|denotes]]`. If no such main note
          exists, then this is `None`.
        - The fourth entry is the MarkdownFile consisting of the "main" content of the note,
          which excludes the information given by all of the other entries.
        - The fifth entry is the MarkdownFile consisting of the ending bulleted list, listing
          the notations used in the notation notes along with links to the notation notes
          describing these notations. If there is not such bulleted list, then this entry
          is `None`. 

    **Raises**

    - UserWarning
        - If the (non-YAML frontmatter meta) contents of the note do not start
        inn the form `<Notation> [[<link_to_note>|denotes]]`; the name of the
        notation note is included in the warning message.
    - ValueError
        - If the notation note is not formatted correctly by starting
        with the notation with dollar signs `$`.
    - AssertionError
        - If `notation_note` is not determined to be a notation note.
    """
    if isinstance(notation_note, str):
        notation_note = VaultNote(vault, name=notation_note)
    if not vault:
        vault = notation_note.vault
    assert_note_is_of_type(notation_note, PersonalNoteTypeEnum.NOTATION_NOTE)

    mf = MarkdownFile.from_vault_note(notation_note)
    metadata = mf.metadata()
    mf_without_metadata = MarkdownFile(
        [part for part in mf.parts if part['type'] != MarkdownLineEnum.META])

    file_text = str(mf_without_metadata)

    main_mf, mf_with_links_to_notations = _divide_bulleted_list_mf_at_end(mf_without_metadata)
    _remove_the_notation_str_and_denotes_in_main_mf(main_mf, notation_note)

    return (metadata, _get_notation_string(file_text, notation_note),
            _main_of_notation_from_text(file_text), main_mf,
            mf_with_links_to_notations)


def _get_notation_string(
        file_text: str,
        notation_note: VaultNote
        ) -> str:
    """Return the notation string from the text of the notation note..

    Assumes that the notation string exists and is well formatted.

    Helper function for `parse_notation_note`.
    """
    try:
        return re.search(r'\$.+?\$', file_text).group()
    except AttributeError as e:
        raise ValueError(
            'There seems to be a formatting error in a notation note'
            ' and the notation has not been identified. The following is the'
            f' name of the notation note: {notation_note.name}')
    

def _remove_the_notation_str_and_denotes_in_main_mf(
        main_mf: MarkdownFile,
        notation_note: VaultNote):
    """Remove the text `<notation> denotes ` which starts the
    notation description.

    Helper function of `parse_notation_note`.
    """
    for part in main_mf.parts:
        if part['type'] == MarkdownLineEnum.BLANK_LINE:
            continue
        match = re.match(fr'^\$.+?\$ ({WIKILINK_PATTERN}|denotes)\s*', part['line']) 
        if match is None:
            raise ValueError(
                'There seems to be a formatting error in a notation note'
                ' and the notation has not been identified. The following is the'
                f' name of the notation note: {notation_note.name}')
        else:
            start, end = match.span()
            part['line'] = part['line'][end:]
            break
    

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 20
def notation_in_note(
        notation_note: Union[str, VaultNote],
        vault: Optional[PathLike] = None 
        ) -> str:
    """Return the name of the note from which the notation comes from.
    
    **Parameters**

    - `notation_note` - Union[str, VaultNote]
        - Either
            
            - The name of the notation note or
            - The `VaultNote` object of the notation note. 
            
        The note name is expected to be unique
        inside the vault specified by `vault`. 
        This is expected to contain `'notation'` as a substring. 
        Usually, this is expected to be formatted in one of
        the following forms:
            - `'<reference_name>_notation_<rest_of_note_name>'`
            - `'notation.<rest_of_note_name>'
    - `vault` - Pathlike or `None`
        - Defaults to `None`
        
    **Returns**

    - str
        - The notation in LaTeX, including the dollar signs `$`.

    **Raises**

    """
    _, notation_in_note, _, _, _ = parse_notation_note(notation_note, vault)
    return notation_in_note

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 23
def main_of_notation(
        notation_note: VaultNote, # The VaultNote object representing the notation note.
        as_note: bool = False # If `False`, then returns the name of the note, and returns a VaultNote object with the same vault as `notation_note` otherwise. The vault used to get the `VaultNote` is the vault of `notation_note`.
        ) -> Union[str, VaultNote, None]: # The (name of the) main information note that `notation_note` comes from. Returns `None` if `notation_note` does not come from such a note.
    """Return the name of the note from which the notation comes from.
            
    **Raises**

    - ValueError
        - If the notation note is not formatted correctly by starting
        with the notation with dollar signs `$`.
    """
    vault = notation_note.vault
    _, _, main_note_name, _, _ = parse_notation_note(notation_note, vault)
    if main_note_name is None:
        return None
    if as_note:
        return VaultNote(notation_note.vault, name=main_note_name)
    else:
        return main_note_name

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 32
def notation_str_in_a_standard_information_note(
        info_note: VaultNote
        ) -> list[str]: # Each str is a LaTeX str, beginning and trailing dollar signs `$` (single or double) included.
    """
    Return the LaTeX str's with notations in a standard information note.

    A LaTeX str is deemed to be a notation if it is surrounded by double
    asterisks `**`
    """
    mf = MarkdownFile.from_vault_note(info_note)
    notations = []
    for part in mf.parts:
        indices = notation_asterisk_indices(part['line'])
        notations.extend([
            part['line'][start+2:end-2] for start, end in indices])
    return notations

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 37
def notation_notes_linked_in_see_also_section(
        info_note: VaultNote,
        vault: PathLike, # Path to the vault directory.
        as_vault_notes: bool = True # If `True`, returns each notation note as a `VaultNote` object.  Otherwise, returns the name of each notation note. Defaults to `True`.
        ) -> Union[list[VaultNote], list[str]]: # Each entry corresponds to a notation note in the vault.
    """Return a list of notation notes listed in the
    `See Also` section of the standard information note.

    In the current implementation of this function, only 
    "notation notes" that actually exist are included in
    the returned list.
    """
    links = bulleted_links_of_type_in_section(
        info_note, vault, section="See Also",
        note_type=PersonalNoteTypeEnum.NOTATION_NOTE)
    note_names = [link.file_name for link in links]
    if as_vault_notes:
        return [VaultNote(vault, name=note_name) for note_name in note_names]
    else:
        return note_names


# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 46
def notations_and_main_notes(
        vault: PathLike, # Path to the vault directory.
        subdirectory: Optional[PathLike] = None, # Path to the subdirectory, relative to `vault`, to find the notation notes. Searches for all notation notes here and in subdirectories of this subdirectory. If `None`, then the `note parameter is used to determined the subdirectory. If `subdirectory` is the empty str, then all notation notes in the vault are searched. Defaults to `None`. 
        note: Optional[VaultNote] = None # A note in the vault. The directory that this note is in determines the `subdirectory` parameter if the argument passed to `subdirectory` is the blank str. This note can usually be an index note, e.g. `'_index_silverman'`. Defaults to `None`, in which case `subdirectory` must be specified.
        ) -> dict[str, Union[str, None]]: # A key is the unique name of a notation note in the vault and its corresponding value is the name of the main note of the notation note. Each main note may not actually exist, but each notation note definitely exists. If the notation note has no main note (i.e. has no links to other notes), then the value is `None`.
    """Return a `dict` with all of notation notes in a specified
    subdirectory of a vault and the names of the main notes of these
    notation notes.
    
    **Returns**
    - `dict`

    **Raises**

    - ValueError
        - If `subdirectory` and `note` are both `None`.
    """
    if note is None and subdirectory is None:
        raise ValueError(
            'Both the `subdirectory` and `note` parameters are None.')
    # vault = vault if vault is not None else ''
    if subdirectory is None:
        subdirectory = Path(note.rel_path).parent
    subdirectory_path = Path(vault) / subdirectory
    notes_in_subdirectory = subdirectory_path.glob(f'**/*.md') 
    relative_paths = [Path(note_path).relative_to(subdirectory_path)
                      for note_path in notes_in_subdirectory]
    vn_objects = [VaultNote(vault, rel_path=Path(subdirectory) / rel_path)
                  for rel_path in relative_paths]
    return {vn.name: main_of_notation(vn) for vn in vn_objects
            if note_is_of_type(vn, PersonalNoteTypeEnum.NOTATION_NOTE)}

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 54
def notation_note_is_linked_in_see_also_section(
        notation_note: VaultNote,
        info_note: Optional[VaultNote] = None # The note in which to find the link to `notation_note`. Defaults to `None`, in which case the main note is determined to be the first linked note of `notation_note`.
        ) -> bool:
    """Return `True` if a notation note is linked in the `See Also`
    section of a standard information note. 
    """
    if not info_note:
        info_note = main_of_notation(notation_note, as_note=True)
    notes = notation_notes_linked_in_see_also_section(
        info_note, vault=notation_note.vault, as_vault_notes=False)
    return notation_note.name in notes


# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 61
def add_notation_note_to_see_also(
        notation_note: VaultNote,
        info_note: Optional[VaultNote] = None, # The note in which to link `notation_note`. Defaults to `None`, in which case the main note is determined to be the first linked note of `notation_note`.
        do_not_repeat: bool = True # If `True`, do not add a link to `notation_note` in if there is already a such a link.
        ) -> None:
    """Add a link to a notation note in the `See Also` section of
    a standard information note.

    **Raises**

    - NoteDoesNotExistError
        - If the information note to link to does not exist.
    
    """
    if not info_note:
        info_note = main_of_notation(notation_note, as_note=True)
    if not info_note.exists():
        raise NoteDoesNotExistError
    if do_not_repeat and notation_note_is_linked_in_see_also_section(
            notation_note, info_note):
        return
    mf = MarkdownFile.from_vault_note(info_note)
    link = ObsidianLink(False, notation_note.name, 0, 0)
    mf.add_line_in_section(
        'See Also', {'type': MarkdownLineEnum.UNORDERED_LIST,
                     'line': f'- {str(link)}'})
    mf.write(info_note)



# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 74
def add_missing_notation_links_to_information_notes(
        vault: PathLike, # Path to the vault directory.
        subdirectory: Optional[PathLike] = None, # Path to the subdirectory, relative to `vault`, to find the notation notes and their main notes. Searches for all notation notes here and in subdirectories of this subdirectory. If `None`, then the `note` parameter is used to determine `subdirectory`. Defaults to `None`. 
        note: Optional[VaultNote] = None # A note in the vault. The directory that this note is in determines the `subdirectory` parameter if it is `None`.  Defaults to `None`, in which case `subdirectory` must be specified.
        ) -> None:
    # TODO: deal with possibility that note does not exist.
    """For each notation note in a specified subdirectory, Add links
    to notation notes in their main information notes if the notation
    links are not already present.
    
    **Raises**

    - ValueError
        - If `subdirectory` and `note` are both `None`.
    """
    mains_dict = notations_and_main_notes(vault, subdirectory, note)
    to_check = {key: value for key, value in mains_dict.items()
                if value is not None}
    for notation, main in to_check.items():
        if not main:
            continue
        notation_note = VaultNote(vault, name=notation)
        try:
            main_note = VaultNote(vault, name=main)  # TODO add the subdirectory parameter here appropriately. See Also `notations_and_main_note` 
            add_notation_note_to_see_also(notation_note, main_note)
        except NoteDoesNotExistError:
            continue

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 78
def notations_to_add_in_index(
        vault: PathLike, # Path to the vault directory.
        notation_index_note = VaultNote, # The notation index note in the vault where the notations should be added to.
        subdirectory: Optional[PathLike] = None , # Path to the subdirectory, relative to `vault`, to find the notation notes. Searches for all notation notes here and in subdirectories of this subdirectory. If `None`, then the `note parameter is used to determined the subdirectory. If `subdirectory` is the empty str, then all notation notes in the vault are searched. Defaults to `None`. 
        note: Optional[VaultNote] = None # The directory that this note is in determines the argument to `subdirectory` parameter if it is `None`. Defaults to `None`, in which case `subdirectory` must be specified.
        ) -> list[tuple[str, ObsidianLink]]: # Each tuple in the list consists of the notation str of the notation note (including surrounding dollar signs `$`) and the (nonembedded) ObsidianLink object for a link to the notation note.
    """Returns notations and links of notation notes to that ought to be
    added in the corresponding notation index, i.e. are in the reference
    folder but not linked by the notation index note.

    If a notation note is not properly formatted, e.g. does not have a
    notation, then the notation and link for the notation note will not
    be included.
    
    **Raises**
    - ValueError
        - If `subdirectory` and `note` are both `None`.

    """
    vault = vault if vault is not None else ''
    mains_dict = notations_and_main_notes(vault, subdirectory, note)
    mf_object = MarkdownFile.from_file(notation_index_note.path())
    mf_text = str(mf_object)
    notations_and_links = []
    for notation, _ in mains_dict.items():
        link_object = ObsidianLink(
            is_embedded=True, file_name=notation, anchor=0, custom_text=0,
            link_type=LinkType.WIKILINK)
        link = link_object.to_string()
        try:
            notation_str = notation_in_note(notation, vault)
        except AttributeError:  # When a notation note is incomplete.
            continue  # TODO: print a warning
        if not link in mf_text:
            notations_and_links.append((notation_str, link_object))
    return notations_and_links



# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 81
def index_notation_note_formatted_entry(
        notation_str: str, # The str of the notation, including the surrounding dollar signs `$`.
        link: ObsidianLink # The embedded link to the notation note. 
        ) -> str:
    """Return a str formatted for an index notation note entry.

    It is recommended to pass the outputs of
    `notations_to_add_in_index` to this function.
    """
    return f'### {notation_str}\n- {link.to_string()}'

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 85
def make_a_notation_note(
        main_note: VaultNote, # The note from which the notation originates.
        vault: PathLike,
        notation: str, # The notation typed in latex. May or may not be surrounded by dollar signs
        description: str, # The rest of the text describing notation.
        notation_note_name: str, # The name of the new notation note to be created.
        destination: Optional[PathLike] = None, # The directory to create the new notation note in.  If `None`, then creates the new notation note in the same place as the note specified by `note_name`
        overwrite: bool = False, # If `True`, overwrite file of the same path as the new notation file to be written, if such a file exists.  Otherwise, does nothing. Even if a link to the old notation note exists in `main_note`, a new link will still be added.  Defaults to `False`.
        add_to_main: bool = True # If `True`, adds a link to the notation note in the `See Also` section of the main note.
        ) -> Union[VaultNote, None]: # The newly created notation note. If no note is created, then returns `None`.
    """Make a new notation note, optionally add a link to it in the
    `See Also` section of its main note, returns it.

    The notation note is created in the same directory as the main note.
    The meta of the notation note has a `latex_in_original` section which
    lists the contents of the latex string in the main note from which the
    notation note comes from. This is so that the
    `make_notation_notes_from_double_asts` method can distinguish between
    notations for which a note has been created and for which a note has
    not been created.
    """
    if destination is None:
        destination = main_note.directory(relative=True)
    notation_note = VaultNote(
        vault, rel_path=destination / f'{notation_note_name}.md')
    if not overwrite and notation_note.exists():
        return
    if not notation_note.exists():
        notation_note.create()
    to_print = _full_notation_string(main_note, notation, description)
    # TODO: change this to use VaultNote method
    with open(notation_note.path(), 'w+', encoding='utf8') as notation_file:
        notation_file.write(to_print)
    if add_to_main:
        add_notation_note_to_see_also(notation_note, main_note)
    return notation_note
    

def _full_notation_string(
        main_note: VaultNote, notation: str, description: str) -> str:
    """The full "statement" of a notation.
    
    Says something like "<notation> denotes <description of notation>", e.g.
    "$\dim V$ denotes the dimension of the vector space $V$".
    
    **Parameters**
    - notation - str
        - Notation written in LaTeX.
    - description - str
        - The full description of the notation.
        
    **Returns**
    - str
    """
    raw_notation = f'{_raw_notation(notation)}'
    denote_link = ObsidianLink(False, main_note.name, 0, 'denotes')
    meta_notation = raw_notation.replace('\\', '\\\\')
    return (f'---\ndetect_regex: []\n'
            f'latex_in_original: ["{meta_notation}"]'
            f'\n---\n${raw_notation}$ {str(denote_link)} {description}')


def _raw_notation(notation: str):
    """
    """
    notation = notation.strip()
    notation = notation.strip('$')
    notation = notation.strip()
    return notation


# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 99
def make_notation_notes_from_double_asts(
        main_note: VaultNote, # The standard information note from which the notations are marked with double asterisks
        vault: PathLike, # The name of the reference; the notation note's name will start with `{reference_name}_notation_`.
        reference_name: str,
        destination: Optional[PathLike] = None, # The directory to create the new notation notes in.  If `None`, then creates the new notation note in the same place as the note specified by `note_name`
        overwrite: bool = False, # If `True`, overwrite file of the same path as the new notation file to be written, if such a file exists.  Otherwise, does nothing. Defaults to `False`.
        add_to_main: bool = True # If `True`, adds links to the notation note in the `See Also` section of the main note.
        ) -> list[VaultNote]: # The list of VaultNotes that are newly created/modified.
    """Make notation notes based on double asterisks surrounding LaTeX text in
    a standard information note.

    Notations are deemed to be completely LaTeX text in info notes that
    are surrounded by double asterisks. In basicality, if such a LaTeX
    text (without surrounding dollars signs `$` or `$$`) is listed in
    the `latex_in_original` metadata section of some notation note in the same
    directory as the info note whose main note is the info note in question,
    then a new notation note for that LaTeX text 
    is not created. However, if there are multiple instances of the same
    LaTeX text, then some notation notes may be created so that the number
    of times the LaTeX text appears in the info note is the no more than
    the number of times the LaTeX text appears in `latex_in_original` metadata
    sections of notation notes (in the same directory as the info note whose
    main note is the info note).

    For example, if there is an info note with notations `A`, `A`, `'A'`,
    `'A'`, and `B` and if there is a single notation note in the same
    directory as the info note with two `'A'` and `'A'` entries in its
    `latex_in_original` metadata section, then three notation notes will be
    created: two with `'A'` listed in their `latex_in_original` sections, and
    one with `'B'` listed in its `latex_in_original` section.

    **Raises**

    - Warning
        - If there are notation notes whose main note is determined to
        be to `main_note` and whose notations "excessively cover" those
        in `main_note`, i.e. the notation notes have more notations than
        `main_note` introduces. The main note and the excessive
        notations are printed; the notations are printed instead of the 
        notation notes because the same notation may span either multiple
        or single notation notes.
    """
    # Find notations
    notations = notation_str_in_a_standard_information_note(main_note)
    notations = [_raw_notation(notation) for notation in notations]
    # Get only the notations not already made into notes based on
    # latex_in_original
    all_latex_in_original = _latex_in_original_from_notat_notes_to_main_note(
        vault, main_note)
    notations_to_create = Multiset(notations).difference(all_latex_in_original)
    notations_to_create = list(notations_to_create)
    # Alert of existing notations that should not be there
    excess_notations = all_latex_in_original.difference(Multiset(notations))
    excess_notations = list(excess_notations)
    if excess_notations:
        warnings.warn(
            f"The following note has the following excess notations: "
            f"{main_note.name}, {', '.join(excess_notations)}")
    # Make notations
    return _make_new_notes_from_sifted_double_asts(
        main_note, vault, reference_name, notations_to_create,
        destination, overwrite, add_to_main)
    

def _latex_in_original_in_notat(
        notation_note: VaultNote
        ) -> list[str]:
    """Return the `latex_in_original` metadata section of the notation note.
    
    If the `latex_in_original` metadata section does not exist, then returns
    the list consisting of the notation in the notation note.
    """
    # TODO: test in the case that `latex_in_original` section does not exist
    mf = MarkdownFile.from_vault_note(notation_note)
    metadata = mf.metadata()
    if metadata is not None:
        return metadata.get('latex_in_original',
                            [notation_in_note(notation_note).strip('$')])
    else:
        return [notation_in_note(notation_note)]

    
def _latex_in_original_from_notat_notes_to_main_note(
        vault: PathLike,
        main_note: VaultNote # The info note
        ) -> Multiset:
    """Return a Multiset enumerating the entries of `latex_in_original`
    in the notation notes in the same directory as an info note
    """
    notation_notes_in_folder = notations_and_main_notes(vault, note=main_note)
    notation_notes_of_main_note = [
        VaultNote(vault, name=notation_note) for notation_note, info_note
        in notation_notes_in_folder.items()
        if main_note.name == info_note]

    all_latex_in_original = Multiset()
    for notat_note in notation_notes_of_main_note:
        # TODO: assert that _latex_in_original_in_notat outputs list of str;
        # I came across an example where it outputs a list of a list
        # by virtue of the `latex_in_original` field of a notation note
        # being `[[ Y(S)]]`, when it should have been `["[ Y(S)]"]` instead.
        all_latex_in_original.update(_latex_in_original_in_notat(notat_note))
    return all_latex_in_original


MAX_NOTE_NAME_LENGTH = 80
def _make_new_notes_from_sifted_double_asts(
        main_note: VaultNote, vault: PathLike, reference_name: str,
        notations: list[str], destination: Optional[PathLike],
        overwrite: bool, add_to_main: bool) -> list[VaultNote]:
    # TODO: write a test for making a new note from the notation 
    # `\begin{equation} \label{escape rate} 	G_{F_t}(z,w) = \lim_{n\to\infty} \frac{1}{d^n} \log \| F_t^n(z,w) \|, \end{equation}`
    # in particular, test that file path has been sanitized.
    """
    Actually makes the notation notes found from the double asterisked LaTeX
    str's
    """
    # TODO: test that note names aren't too long.
    new_notes = []
    for notation in reversed(notations):
        notation_note_name = f'{reference_name}_notation_'\
            f'{latex_to_path_accepted_string(notation)}'
        if len(notation_note_name) > MAX_NOTE_NAME_LENGTH:
            notation_note_name = notation_note_name[:MAX_NOTE_NAME_LENGTH]
        notation_note_name = VaultNote.unique_name(
            notation_note_name, vault)
        new_note = make_a_notation_note(
            main_note, vault, notation, '', notation_note_name,
            destination, overwrite, add_to_main)
        if new_note:
            new_notes.append(new_note)
    return new_notes
    



# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 118
SPECIAL_CHARACTERS = ['.', '+', '*', '?', '^', '$', '(', ')',
                      '[', ']', '{', '}', '|', '\\']
replaceable_groups = [['mathrm', 'operatorname', 'rm', 'text'],
                      ['mathbf', 'bf'],
                      ['mathit', 'it']]


def _build_replacables_from_groups(
        replaceable_groups: list[list[str]]) -> dict[str, set[str]]:
    total_dict = {}
    for listy in replaceable_groups:
        set_for_group = set(listy)
        for macro in listy:
            total_dict[macro] = set_for_group
    return total_dict


REPLACEABLES = _build_replacables_from_groups(replaceable_groups)
    

def regex_from_latex(
        latex: str, replaceables: dict[str, set[str]] = REPLACEABLES,
        special_characters: list[str] = SPECIAL_CHARACTERS) -> str:
    """Returns regex to match latex math mode string which is essentially
    equivalent to a specified latex math mode string.
    
    The outputs of this function may not work correctly.
    The regex pattern does not have to fully match equivalent string.
    
    **Parameters**

    - latex - str
        - The latex math mode string. Does not include math mode delimiters
        such as `$`, `$$`, `\[ \]` (although the characters `'\['` and `'\]'`
        can still be part of the string, e.g. for optional arguments of a
        macro/operator). Can include "placeholders" `r'\1'`, `r'\2'`, `r'\3'`,
        etc. to indicate substitutable/generics; the placeholders can be
        substituted with any string.
    - replaceables - dict[str, set[str]]
        - latex strings/commands which are considered "interreplacable"
    - special_characters - list[str]
        - characters to add a backslash `'\'` in front of for regex.
        Defaults to a list consisting of special characters in regex.
    """
    if not replaceables:
        replaceables = {}
    w = LatexWalker(latex)
    nodelist, _, _ = w.get_latex_nodes(pos=0)
    regex_parts = []
    # print(nodelist)
    for node in nodelist:
        _look_into_node(node, regex_parts,
                        replaceables, special_characters)
    regex_parts.append('(?:[ \\{\\}]*)')
    return ''.join(regex_parts)
    
def _look_into_node(
        node: LatexNode, regex_parts: list[str],
        replaceables: dict[str, set[str]],
        special_characters: list[str]) -> None:
    """Appends to `regex_parts`"""
    # hasattr(node, 'nodeargd')
    # print(node)
    if isinstance(node, LatexMacroNode):
        macroname = node.macroname
        if _macro_is_actually_placeholder(macroname):
            regex_parts.append('(?:.*)')
        else:
            if macroname in replaceables:
                replaceable_macros = replaceables[macroname]
            else:
                replaceable_macros = [macroname]
            options_str = '|'.join(replaceable_macros)
            options_str = f'(?:{options_str})'
            regex_parts.append(fr'(?: *?)\\{options_str}(?: *?)')
        for node in node.nodeargd.argnlist:
            _look_into_node(node, regex_parts,
                            replaceables, special_characters)
    elif isinstance(node, LatexGroupNode):
        # print('\nGroup Node')
        # print(node)
        # print(node.nodelist)
        delimiters = node.delimiters
        regex_parts.append(f'\\{delimiters[0]}(?: *?)')
        for node in node.nodelist:
            _look_into_node(node, regex_parts,
                            replaceables, special_characters)
        regex_parts.append(f'(?: *?)\\{delimiters[1]}')
    elif isinstance(node, LatexCharsNode):
        # print('\nChars Node')
        # print(node)
        # print(node.chars)
        chars = node.chars.strip()
        chars = list(chars)
        chars = [f'\\{char}' if char in special_characters else char
                 for char in chars]
        # print(chars)
        chars.insert(0, '')  # add the misc spaces/brackets front and back
        chars.append('')
        regex_optional_spaces_and_brackets = '(?:[ \\{\\}]*?)'.join(chars)
        regex_parts.append(regex_optional_spaces_and_brackets)
        
def _macro_is_actually_placeholder(macro: str) -> bool:
    return macro.isnumeric()

# %% ../../../../nbs/20_markdown.obsidian.personal.notation.ipynb 122
def regex_from_notation_note(vault: PathLike, note: VaultNote) -> str:
    """Returns a regex str to detect the notation of the notation note.
    
    The regex detection strings should be in a list labeled `detect_regex` in
    the yaml frontmatter. If multiple strings are in the list, then the regex
    will detect latex math mode strings roughly corresponding to any of them.
    If multiple strings are in the list, then they must be ordered 
    "by priority", with the higher priority regexes coming first. It is good
    to have these string in quotes `""` to make sure that yaml can load them
    safely. When doing so, make sure to escape characters, e.g. backslash
    should be typed as `\\`, etc.
    
    The strings in `detect_regex` can include placeholders, cf.
    ``regex_from_latex``.
    
    **Parameters**
    - vault - PathLike
    - note - VaultNote
    
    **Returns**
    - str
        - Of the regex used to detect the notation. The regex does not need to
        fully match instances of the notation.
    """
    assert note_is_of_type(note, PersonalNoteTypeEnum.NOTATION_NOTE)
    mf = MarkdownFile.from_vault_note(note)
    metadata = mf.metadata()
    if metadata and 'detect_regex' in metadata:
        detects = metadata['detect_regex']
        regexes = [regex_from_latex(detect) for detect in detects]
        return '|'.join(regexes)
    else:
        notation = notation_in_note(note, vault)
        return regex_from_latex(notation[1:-1])  # Get rid of `'$'`.
